# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
# http://code.google.com/p/protobuf/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO(robinson): We probably need to provide deep-copy methods for
# descriptor types.  When a FieldDescriptor is passed into
# Descriptor.__init__(), we should make a deep copy and then set
# containing_type on it.  Alternatively, we could just get
# rid of containing_type (iit's not needed for reflection.py, at least).
#
# TODO(robinson): Print method?
#
# TODO(robinson): Useful __repr__?

"""Descriptors essentially contain exactly the information found in a .proto
file, in types that make this information accessible in Python.
"""

__author__ = 'robinson@google.com (Will Robinson)'


class Error(Exception):
  """Base error for this module."""


class DescriptorBase(object):

  """Descriptors base class.

  This class is the base of all descriptor classes. It provides common options
  related functionaility.

  Attributes:
    has_options:  True if the descriptor has non-default options.  Usually it
        is not necessary to read this -- just call GetOptions() which will
        happily return the default instance.  However, it's sometimes useful
        for efficiency, and also useful inside the protobuf implementation to
        avoid some bootstrapping issues.
  """

  def __init__(self, options, options_class_name):
    """Initialize the descriptor given its options message and the name of the
    class of the options message. The name of the class is required in case
    the options message is None and has to be created.
    """
    self._options = options
    self._options_class_name = options_class_name

    # Does this descriptor have non-default options?
    self.has_options = options is not None

  def GetOptions(self):
    """Retrieves descriptor options.

    This method returns the options set or creates the default options for the
    descriptor.
    """
    if self._options:
      return self._options
    from google.protobuf import descriptor_pb2
    try:
      options_class = getattr(descriptor_pb2, self._options_class_name)
    except AttributeError:
      raise RuntimeError('Unknown options class name %s!' %
                         (self._options_class_name))
    self._options = options_class()
    return self._options


class _NestedDescriptorBase(DescriptorBase):
  """Common class for descriptors that can be nested."""

  def __init__(self, options, options_class_name, name, full_name,
               file, containing_type, serialized_start=None,
               serialized_end=None):
    """Constructor.

    Args:
      options: Protocol message options or None
        to use default message options.
      options_class_name: (str) The class name of the above options.

      name: (str) Name of this protocol message type.
      full_name: (str) Fully-qualified name of this protocol message type,
        which will include protocol "package" name and the name of any
        enclosing types.
      file: (FileDescriptor) Reference to file info.
      containing_type: if provided, this is a nested descriptor, with this
        descriptor as parent, otherwise None.
      serialized_start: The start index (inclusive) in block in the
        file.serialized_pb that describes this descriptor.
      serialized_end: The end index (exclusive) in block in the
        file.serialized_pb that describes this descriptor.
    """
    super(_NestedDescriptorBase, self).__init__(
        options, options_class_name)

    self.name = name
    # TODO(falk): Add function to calculate full_name instead of having it in
    #             memory?
    self.full_name = full_name
    self.file = file
    self.containing_type = containing_type

    self._serialized_start = serialized_start
    self._serialized_end = serialized_end

  def GetTopLevelContainingType(self):
    """Returns the root if this is a nested type, or itself if its the root."""
    desc = self
    while desc.containing_type is not None:
      desc = desc.containing_type
    return desc

  def CopyToProto(self, proto):
    """Copies this to the matching proto in descriptor_pb2.

    Args:
      proto: An empty proto instance from descriptor_pb2.

    Raises:
      Error: If self couldnt be serialized, due to to few constructor arguments.
    """
    if (self.file is not None and
        self._serialized_start is not None and
        self._serialized_end is not None):
      proto.ParseFromString(self.file.serialized_pb[
          self._serialized_start:self._serialized_end])
    else:
      raise Error('Descriptor does not contain serialization.')


class Descriptor(_NestedDescriptorBase):

  """Descriptor for a protocol message type.

  A Descriptor instance has the following attributes:

    name: (str) Name of this protocol message type.
    full_name: (str) Fully-qualified name of this protocol message type,
      which will include protocol "package" name and the name of any
      enclosing types.

    containing_type: (Descriptor) Reference to the descriptor of the
      type containing us, or None if this is top-level.

    fields: (list of FieldDescriptors) Field descriptors for all
      fields in this type.
    fields_by_number: (dict int -> FieldDescriptor) Same FieldDescriptor
      objects as in |fields|, but indexed by "number" attribute in each
      FieldDescriptor.
    fields_by_name: (dict str -> FieldDescriptor) Same FieldDescriptor
      objects as in |fields|, but indexed by "name" attribute in each
      FieldDescriptor.

    nested_types: (list of Descriptors) Descriptor references
      for all protocol message types nested within this one.
    nested_types_by_name: (dict str -> Descriptor) Same Descriptor
      objects as in |nested_types|, but indexed by "name" attribute
      in each Descriptor.

    enum_types: (list of EnumDescriptors) EnumDescriptor references
      for all enums contained within this type.
    enum_types_by_name: (dict str ->EnumDescriptor) Same EnumDescriptor
      objects as in |enum_types|, but indexed by "name" attribute
      in each EnumDescriptor.
    enum_values_by_name: (dict str -> EnumValueDescriptor) Dict mapping
      from enum value name to EnumValueDescriptor for that value.

    extensions: (list of FieldDescriptor) All extensions defined directly
      within this message type (NOT within a nested type).
    extensions_by_name: (dict, string -> FieldDescriptor) Same FieldDescriptor
      objects as |extensions|, but indexed by "name" attribute of each
      FieldDescriptor.

    is_extendable:  Does this type define any extension ranges?

    options: (descriptor_pb2.MessageOptions) Protocol message options or None
      to use default message options.

    file: (FileDescriptor) Reference to file descriptor.
  """

  def __init__(self, name, full_name, filename, containing_type, fields,
               nested_types, enum_types, extensions, options=None,
               is_extendable=True, extension_ranges=None, file=None,
               serialized_start=None, serialized_end=None):
    """Arguments to __init__() are as described in the description
    of Descriptor fields above.

    Note that filename is an obsolete argument, that is not used anymore.
    Please use file.name to access this as an attribute.
    """
    super(Descriptor, self).__init__(
        options, 'MessageOptions', name, full_name, file,
        containing_type, serialized_start=serialized_start,
        serialized_end=serialized_start)

    # We have fields in addition to fields_by_name and fields_by_number,
    # so that:
    #   1. Clients can index fields by "order in which they're listed."
    #   2. Clients can easily iterate over all fields with the terse
    #      syntax: for f in descriptor.fields: ...
    self.fields = fields
    for field in self.fields:
      field.containing_type = self
    self.fields_by_number = dict((f.number, f) for f in fields)
    self.fields_by_name = dict((f.name, f) for f in fields)

    self.nested_types = nested_types
    self.nested_types_by_name = dict((t.name, t) for t in nested_types)

    self.enum_types = enum_types
    for enum_type in self.enum_types:
      enum_type.containing_type = self
    self.enum_types_by_name = dict((t.name, t) for t in enum_types)
    self.enum_values_by_name = dict(
        (v.name, v) for t in enum_types for v in t.values)

    self.extensions = extensions
    for extension in self.extensions:
      extension.extension_scope = self
    self.extensions_by_name = dict((f.name, f) for f in extensions)
    self.is_extendable = is_extendable
    self.extension_ranges = extension_ranges

    self._serialized_start = serialized_start
    self._serialized_end = serialized_end

  def CopyToProto(self, proto):
    """Copies this to a descriptor_pb2.DescriptorProto.

    Args:
      proto: An empty descriptor_pb2.DescriptorProto.
    """
    # This function is overriden to give a better doc comment.
    super(Descriptor, self).CopyToProto(proto)


# TODO(robinson): We should have aggressive checking here,
# for example:
#   * If you specify a repeated field, you should not be allowed
#     to specify a default value.
#   * [Other examples here as needed].
#
# TODO(robinson): for this and other *Descriptor classes, we
# might also want to lock things down aggressively (e.g.,
# prevent clients from setting the attributes).  Having
# stronger invariants here in general will reduce the number
# of runtime checks we must do in reflection.py...
class FieldDescriptor(DescriptorBase):

  """Descriptor for a single field in a .proto file.

  A FieldDescriptor instance has the following attriubtes:

    name: (str) Name of this field, exactly as it appears in .proto.
    full_name: (str) Name of this field, including containing scope.  This is
      particularly relevant for extensions.
    index: (int) Dense, 0-indexed index giving the order that this
      field textually appears within its message in the .proto file.
    number: (int) Tag number declared for this field in the .proto file.

    type: (One of the TYPE_* constants below) Declared type.
    cpp_type: (One of the CPPTYPE_* constants below) C++ type used to
      represent this field.

    label: (One of the LABEL_* constants below) Tells whether this
      field is optional, required, or repeated.
    has_default_value: (bool) True if this field has a default value defined,
      otherwise false.
    default_value: (Varies) Default value of this field.  Only
      meaningful for non-repeated scalar fields.  Repeated fields
      should always set this to [], and non-repeated composite
      fields should always set this to None.

    containing_type: (Descriptor) Descriptor of the protocol message
      type that contains this field.  Set by the Descriptor constructor
      if we're passed into one.
      Somewhat confusingly, for extension fields, this is the
      descriptor of the EXTENDED message, not the descriptor
      of the message containing this field.  (See is_extension and
      extension_scope below).
    message_type: (Descriptor) If a composite field, a descriptor
      of the message type contained in this field.  Otherwise, this is None.
    enum_type: (EnumDescriptor) If this field contains an enum, a
      descriptor of that enum.  Otherwise, this is None.

    is_extension: True iff this describes an extension field.
    extension_scope: (Descriptor) Only meaningful if is_extension is True.
      Gives the message that immediately contains this extension field.
      Will be None iff we're a top-level (file-level) extension field.

    options: (descriptor_pb2.FieldOptions) Protocol message field options or
      None to use default field options.
  """

  # Must be consistent with C++ FieldDescriptor::Type enum in
  # descriptor.h.
  #
  # TODO(robinson): Find a way to eliminate this repetition.
  TYPE_DOUBLE         = 1
  TYPE_FLOAT          = 2
  TYPE_INT64          = 3
  TYPE_UINT64         = 4
  TYPE_INT32          = 5
  TYPE_FIXED64        = 6
  TYPE_FIXED32        = 7
  TYPE_BOOL           = 8
  TYPE_STRING         = 9
  TYPE_GROUP          = 10
  TYPE_MESSAGE        = 11
  TYPE_BYTES          = 12
  TYPE_UINT32         = 13
  TYPE_ENUM           = 14
  TYPE_SFIXED32       = 15
  TYPE_SFIXED64       = 16
  TYPE_SINT32         = 17
  TYPE_SINT64         = 18
  MAX_TYPE            = 18

  # Must be consistent with C++ FieldDescriptor::CppType enum in
  # descriptor.h.
  #
  # TODO(robinson): Find a way to eliminate this repetition.
  CPPTYPE_INT32       = 1
  CPPTYPE_INT64       = 2
  CPPTYPE_UINT32      = 3
  CPPTYPE_UINT64      = 4
  CPPTYPE_DOUBLE      = 5
  CPPTYPE_FLOAT       = 6
  CPPTYPE_BOOL        = 7
  CPPTYPE_ENUM        = 8
  CPPTYPE_STRING      = 9
  CPPTYPE_MESSAGE     = 10
  MAX_CPPTYPE         = 10

  # Must be consistent with C++ FieldDescriptor::Label enum in
  # descriptor.h.
  #
  # TODO(robinson): Find a way to eliminate this repetition.
  LABEL_OPTIONAL      = 1
  LABEL_REQUIRED      = 2
  LABEL_REPEATED      = 3
  MAX_LABEL           = 3

  def __init__(self, name, full_name, index, number, type, cpp_type, label,
               default_value, message_type, enum_type, containing_type,
               is_extension, extension_scope, options=None,
               has_default_value=True):
    """The arguments are as described in the description of FieldDescriptor
    attributes above.

    Note that containing_type may be None, and may be set later if necessary
    (to deal with circular references between message types, for example).
    Likewise for extension_scope.
    """
    super(FieldDescriptor, self).__init__(options, 'FieldOptions')
    self.name = name
    self.full_name = full_name
    self.index = index
    self.number = number
    self.type = type
    self.cpp_type = cpp_type
    self.label = label
    self.has_default_value = has_default_value
    self.default_value = default_value
    self.containing_type = containing_type
    self.message_type = message_type
    self.enum_type = enum_type
    self.is_extension = is_extension
    self.extension_scope = extension_scope


class EnumDescriptor(_NestedDescriptorBase):

  """Descriptor for an enum defined in a .proto file.

  An EnumDescriptor instance has the following attributes:

    name: (str) Name of the enum type.
    full_name: (str) Full name of the type, including package name
      and any enclosing type(s).

    values: (list of EnumValueDescriptors) List of the values
      in this enum.
    values_by_name: (dict str -> EnumValueDescriptor) Same as |values|,
      but indexed by the "name" field of each EnumValueDescriptor.
    values_by_number: (dict int -> EnumValueDescriptor) Same as |values|,
      but indexed by the "number" field of each EnumValueDescriptor.
    containing_type: (Descriptor) Descriptor of the immediate containing
      type of this enum, or None if this is an enum defined at the
      top level in a .proto file.  Set by Descriptor's constructor
      if we're passed into one.
    file: (FileDescriptor) Reference to file descriptor.
    options: (descriptor_pb2.EnumOptions) Enum options message or
      None to use default enum options.
  """

  def __init__(self, name, full_name, filename, values,
               containing_type=None, options=None, file=None,
               serialized_start=None, serialized_end=None):
    """Arguments are as described in the attribute description above.

    Note that filename is an obsolete argument, that is not used anymore.
    Please use file.name to access this as an attribute.
    """
    super(EnumDescriptor, self).__init__(
        options, 'EnumOptions', name, full_name, file,
        containing_type, serialized_start=serialized_start,
        serialized_end=serialized_start)

    self.values = values
    for value in self.values:
      value.type = self
    self.values_by_name = dict((v.name, v) for v in values)
    self.values_by_number = dict((v.number, v) for v in values)

    self._serialized_start = serialized_start
    self._serialized_end = serialized_end

  def CopyToProto(self, proto):
    """Copies this to a descriptor_pb2.EnumDescriptorProto.

    Args:
      proto: An empty descriptor_pb2.EnumDescriptorProto.
    """
    # This function is overriden to give a better doc comment.
    super(EnumDescriptor, self).CopyToProto(proto)


class EnumValueDescriptor(DescriptorBase):

  """Descriptor for a single value within an enum.

    name: (str) Name of this value.
    index: (int) Dense, 0-indexed index giving the order that this
      value appears textually within its enum in the .proto file.
    number: (int) Actual number assigned to this enum value.
    type: (EnumDescriptor) EnumDescriptor to which this value
      belongs.  Set by EnumDescriptor's constructor if we're
      passed into one.
    options: (descriptor_pb2.EnumValueOptions) Enum value options message or
      None to use default enum value options options.
  """

  def __init__(self, name, index, number, type=None, options=None):
    """Arguments are as described in the attribute description above."""
    super(EnumValueDescriptor, self).__init__(options, 'EnumValueOptions')
    self.name = name
    self.index = index
    self.number = number
    self.type = type


class ServiceDescriptor(_NestedDescriptorBase):

  """Descriptor for a service.

    name: (str) Name of the service.
    full_name: (str) Full name of the service, including package name.
    index: (int) 0-indexed index giving the order that this services
      definition appears withing the .proto file.
    methods: (list of MethodDescriptor) List of methods provided by this
      service.
    options: (descriptor_pb2.ServiceOptions) Service options message or
      None to use default service options.
    file: (FileDescriptor) Reference to file info.
  """

  def __init__(self, name, full_name, index, methods, options=None, file=None,
               serialized_start=None, serialized_end=None):
    super(ServiceDescriptor, self).__init__(
        options, 'ServiceOptions', name, full_name, file,
        None, serialized_start=serialized_start,
        serialized_end=serialized_end)
    self.index = index
    self.methods = methods
    # Set the containing service for each method in this service.
    for method in self.methods:
      method.containing_service = self

  def FindMethodByName(self, name):
    """Searches for the specified method, and returns its descriptor."""
    for method in self.methods:
      if name == method.name:
        return method
    return None

  def CopyToProto(self, proto):
    """Copies this to a descriptor_pb2.ServiceDescriptorProto.

    Args:
      proto: An empty descriptor_pb2.ServiceDescriptorProto.
    """
    # This function is overriden to give a better doc comment.
    super(ServiceDescriptor, self).CopyToProto(proto)


class MethodDescriptor(DescriptorBase):

  """Descriptor for a method in a service.

  name: (str) Name of the method within the service.
  full_name: (str) Full name of method.
  index: (int) 0-indexed index of the method inside the service.
  containing_service: (ServiceDescriptor) The service that contains this
    method.
  input_type: The descriptor of the message that this method accepts.
  output_type: The descriptor of the message that this method returns.
  options: (descriptor_pb2.MethodOptions) Method options message or
    None to use default method options.
  """

  def __init__(self, name, full_name, index, containing_service,
               input_type, output_type, options=None):
    """The arguments are as described in the description of MethodDescriptor
    attributes above.

    Note that containing_service may be None, and may be set later if necessary.
    """
    super(MethodDescriptor, self).__init__(options, 'MethodOptions')
    self.name = name
    self.full_name = full_name
    self.index = index
    self.containing_service = containing_service
    self.input_type = input_type
    self.output_type = output_type


class FileDescriptor(DescriptorBase):
  """Descriptor for a file. Mimics the descriptor_pb2.FileDescriptorProto.

  name: name of file, relative to root of source tree.
  package: name of the package
  serialized_pb: (str) Byte string of serialized
    descriptor_pb2.FileDescriptorProto.
  """

  def __init__(self, name, package, options=None, serialized_pb=None):
    """Constructor."""
    super(FileDescriptor, self).__init__(options, 'FileOptions')

    self.name = name
    self.package = package
    self.serialized_pb = serialized_pb

  def CopyToProto(self, proto):
    """Copies this to a descriptor_pb2.FileDescriptorProto.

    Args:
      proto: An empty descriptor_pb2.FileDescriptorProto.
    """
    proto.ParseFromString(self.serialized_pb)


def _ParseOptions(message, string):
  """Parses serialized options.

  This helper function is used to parse serialized options in generated
  proto2 files. It must not be used outside proto2.
  """
  message.ParseFromString(string)
  return message



#!/usr/bin/env python
#
# Copyright 2003-2009 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegbot Core Application.

This is the Kegbot Core application, which runs the main drink recording and
post-processing loop. There is exactly one instance of a kegbot core per kegbot
system.

For more information, please see the kegbot documentation.
"""

import time

import gflags

from pykeg.core import alarm
from pykeg.core import backend
from pykeg.core import event
from pykeg.core import kb_app
from pykeg.core import kb_threads
from pykeg.core import manager
from pykeg.core.net import kegnet
from pykeg.core.net import kegnet_pb2

FLAGS = gflags.FLAGS

class KegbotEnv(object):
  """ A class that wraps the context of the kegbot core.

  An instance of this class owns all the threads and services used in the kegbot
  core. It is commonly passed around to objects that the core creates.
  """
  def __init__(self):
    self._event_hub = event.EventHub()

    self._kegnet_server = kegnet.KegnetServer(name='kegnet', kb_env=self,
        addr=FLAGS.kb_core_bind_addr)

    self._backend = backend.KegbotBackend()

    # Build managers
    self._alarm_manager = alarm.AlarmManager()
    self._tap_manager = manager.TapManager('tap-manager', self._event_hub)
    self._flow_manager = manager.FlowManager('flow-manager', self._event_hub,
        self._tap_manager)
    self._authentication_manager = manager.AuthenticationManager('auth-manager',
        self._event_hub, self._flow_manager, self._tap_manager, self._backend)
    self._drink_manager = manager.DrinkManager('drink-manager', self._event_hub,
        self._backend)
    self._thermo_manager = manager.ThermoManager('thermo-manager',
        self._event_hub, self._backend)
    self._token_manager = manager.TokenManager('token-manager', self._event_hub,
        self._backend)
    self._subscription_manager = manager.SubscriptionManager('pubsub',
        self._event_hub, self._kegnet_server)

    # Build threads
    self._threads = set()
    self._service_thread = kb_threads.EventHandlerThread(self, 'service-thread')
    self._service_thread.AddEventHandler(self._tap_manager)
    self._service_thread.AddEventHandler(self._flow_manager)
    self._service_thread.AddEventHandler(self._drink_manager)
    self._service_thread.AddEventHandler(self._thermo_manager)
    self._service_thread.AddEventHandler(self._authentication_manager)
    self._service_thread.AddEventHandler(self._token_manager)
    self._service_thread.AddEventHandler(self._subscription_manager)

    self.AddThread(self._service_thread)

    self.AddThread(kb_threads.EventHubServiceThread(self, 'eventhub-thread'))
    self.AddThread(kb_threads.NetProtocolThread(self, 'net-thread'))
    self.AddThread(kb_threads.AlarmManagerThread(self, 'alarmmanager-thread'))
    self.AddThread(kb_threads.HeartbeatThread(self, 'heartbeat-thread'))

    self._watchdog_thread = kb_threads.WatchdogThread(self, 'watchdog-thread')
    self.AddThread(self._watchdog_thread)

    for tap in self._backend.GetAllTaps():
      # TODO: get rid of max_tick_delta parameter entirely
      self._tap_manager.RegisterTap(tap.meter_name, tap.ml_per_tick,
          (1/tap.ml_per_tick*500))

  def AddThread(self, thr):
    self._threads.add(thr)
    if isinstance(thr, kb_threads.CoreThread):
      self.GetEventHub().AddListener(thr)

  def GetWatchdogThread(self):
    return self._watchdog_thread

  def GetAlarmManager(self):
    return self._alarm_manager

  def GetBackend(self):
    return self._backend

  def GetKegnetServer(self):
    return self._kegnet_server

  def GetEventHub(self):
    return self._event_hub

  def GetTapManager(self):
    return self._tap_manager

  def GetFlowManager(self):
    return self._flow_manager

  def GetAuthenticationManager(self):
    return self._authentication_manager

  def GetTokenManager(self):
    return self._token_manager

  def GetThreads(self):
    return self._threads


class KegbotCoreApp(kb_app.App):
  def __init__(self, name='core'):
    kb_app.App.__init__(self, name)
    self._env = KegbotEnv()

  def _MainLoop(self):
    event = kegnet_pb2.StartCompleteEvent()
    self._env.GetEventHub().PublishEvent(event)
    watchdog = self._env.GetWatchdogThread()
    while not self._do_quit:
      try:
        watchdog.join(1)
        if not watchdog.isAlive():
          self._logger.error("Watchdog thread exited, quitting")
          self.Quit()
          return
      except KeyboardInterrupt:
        self._logger.info("Got keyboard interrupt, quitting")
        self.Quit()
        return

  def _Setup(self):
    kb_app.App._Setup(self)
    for thr in self._env.GetThreads():
      self._AddAppThread(thr)

  def Quit(self):
    self._do_quit = True
    event = kegnet_pb2.QuitEvent()
    self._env.GetEventHub().PublishEvent(event)
    time.sleep(1.0)

    self._logger.info('Stopping any remaining threads')
    self._StopThreads()
    self._logger.info('Kegbot stopped.')
    self._TeardownLogging()




#!/usr/bin/env python
#
# Copyright 2008 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions for use by the views module."""

from pykeg.core import models
from pykeg.core import units

def keg_drinkers_by_volume(keg):
  return drinkers_by_volume(keg.drinks.valid())

def drinkers_by_volume(drinks):
  ret = {}
  for d in drinks:
    if not d.user:
      continue
    ret[d.user] = ret.get(d.user, units.Quantity(0, units.RECORD_UNIT)) + d.Volume()
  outlist = []
  for user, totalvol in ret.iteritems():
    if not user.is_active:
      continue
    outlist.append((totalvol, user))
  outlist.sort(reverse=True)
  return outlist

def drinkers_by_cost(drinks):
  cost_map = {}
  for d in drinks:
    if not d.user:
      continue
    keg_volume = d.keg.size.Volume()
    keg_cost = d.keg.origcost
    drink_pct_keg = d.Volume() / keg_volume
    drink_cost = keg_cost * drink_pct_keg
    cost_map[d.user] = cost_map.get(d.user, 0.0) + drink_cost
  ret = [(v, k) for (k, v) in cost_map.iteritems()]
  ret.sort(reverse=True)
  return ret



import datetime

try:
  import pygooglechart
  HAVE_GOOGLE_CHART = True
except ImportError:
  HAVE_GOOGLE_CHART = False

from django.core import urlresolvers
from django.template import Library
from django.template import Node
from django.template import TemplateSyntaxError
from django.template import Variable
from django.utils.safestring import mark_safe

from pykeg.core import models
from pykeg.core import units

class KegweblibError(Exception):
  """Base kegweblib execption."""

class ChartUnavailableError(KegweblibError):
  """Thrown when a chart cannot be rendered."""

register = Library()

def mugshot_box(user, boxsize=100):
  if user:
    img_url = user.get_profile().MugshotUrl()
    user_url = urlresolvers.reverse('drinker', args=(user.username,))
  else:
    img_url = urlresolvers.reverse('site-media',
        args=('images/unknown-drinker.png',))
    user_url = ''

  return {
      'user' : user,
      'boxsize' : boxsize,
      'user_url' : user_url,
      'img_url': img_url,
  }
register.inclusion_tag('kegweb/mugshot_box.html')(mugshot_box)

def render_page(page):
  return {'page' : page}
register.inclusion_tag('kegweb/page_block.html')(render_page)

def timeago(parser, token):
  """{% timeago <timestamp> %}"""
  tokens = token.contents.split()
  if len(tokens) != 2:
    raise TemplateSyntaxError, '%s requires 2 tokens' % tokens[0]
  return TimeagoNode(tokens[1])
register.tag('timeago', timeago)

class TimeagoNode(Node):
  def __init__(self, timestamp_varname):
    self._timestamp_varname = timestamp_varname

  def render(self, context):
    tv = Variable(self._timestamp_varname)
    ts = tv.resolve(context)

    iso = ts.isoformat()
    alt = ts.strftime("%A, %B %d, %Y %I:%M%p")
    return '<abbr class="timeago" title="%s">%s</abbr>' % (iso, alt)

def chart(parser, tokens):
  """{% chart <charttype> [args] width height %}"""
  tokens = tokens.contents.split()
  if len(tokens) < 4:
    raise TemplateSyntaxError('chart requires at least 4 arguments')
  charttype = tokens[1]
  try:
    width = int(tokens[-2])
    height = int(tokens[-1])
  except ValueError:
    raise TemplateSyntaxError('invalid width or height')
  args = tokens[2:-2]
  return ChartNode(charttype, width, height, args)

register.tag('chart', chart)

class ChartNode(Node):
  def __init__(self, charttype, width, height, args):
    self._charttype = charttype
    self._width = width
    self._height = height
    self._args = args

    self._chart_fn = getattr(self, 'chart_%s' % (self._charttype,))
    if not self._chart_fn:
      raise TemplateSyntaxError('unknown chart type: %s' % self._charttype)

  def render(self, context):
    try:
      if not HAVE_GOOGLE_CHART:
        raise ChartUnavailableError, "chart library unavailable"
      url = self._chart_fn(context)
      res = '<div class="kb-chart"><img src="%s"/></div>' % url
    except ChartUnavailableError, e:
      reason = str(e)
      res = ('<div class="kb-chart-missing" '
          'style="width:%ipx; height:%ipx;">%s'
          '</div>' % (self._width, self._height, reason))
    return res

  def chart_sensor(self, context):
    """ Shows a simple line plot of a specific temperature sensor.

    Syntax:
      {% chart sensor <sensorname> width height %}
    Args:
      sensorname - the nice_name of a ThermoSensor
    """
    sensor_name_var = Variable(self._args[0])
    sensor_name = sensor_name_var.resolve(context)

    try:
      sensor = models.ThermoSensor.objects.get(nice_name=sensor_name)
    except models.ThermoSensor.DoesNotExist:
      raise ChartUnavailableError, "Sensor '%s' not found" % sensor_name

    hours = 12
    now = datetime.datetime.now()
    start = now - (datetime.timedelta(hours=hours))
    start = start - (datetime.timedelta(seconds=start.second))

    points = sensor.thermolog_set.filter(time__gte=start).order_by('time')

    curr = start
    temps = []
    have_temps = False
    for point in points:
      while curr <= point.time:
        curr += datetime.timedelta(minutes=1)
        if curr < point.time:
          temps.append(None)
        else:
          temps.append(point.temp)
          have_temps = True

    if not have_temps:
      raise ChartUnavailableError, "Not enough data"

    chart = pygooglechart.SparkLineChart(self._width, self._height)
    chart.add_data(temps)
    chart.fill_solid(pygooglechart.Chart.BACKGROUND, '00000000')
    chart.set_colours(['4D89F9'])
    return chart.get_url()

  def _get_keg_stats(self, context):
    keg_var = Variable(self._args[0])
    keg = keg_var.resolve(context)

    if not keg or not isinstance(keg, models.Keg):
      raise ChartUnavailableError, "Argument is not a Keg instance"

    stats = keg.GetStats()
    if not stats:
      raise ChartUnavailableError, "Stats unavailable"
    return keg, stats

  def chart_keg_volume(self, context):
    """ Shows a horizontal bar chart of keg served/remaining volume.

    Syntax:
      {% chart keg_volume <keg> width height %}
    Args:
      keg - the keg instance to chart
    """
    keg, stats = self._get_keg_stats(context)

    served = units.Quantity(stats.get('total_volume', 0.0))
    served_pints = served.ConvertTo.Pint
    full_pints = keg.full_volume().ConvertTo.Pint
    remain_pints = full_pints - served_pints

    chart = pygooglechart.StackedHorizontalBarChart(self._width, self._height,
        x_range=(0, full_pints))
    chart.set_bar_width(20)
    chart.set_colours(['4D89F9'])
    chart.add_data([served_pints])
    return chart.get_url()

  def chart_keg_volume_by_day(self, context):
    """ Shows a horizontal bar chart of keg served/remaining volume.

    Syntax:
      {% chart keg_volume_by_day <keg> width height %}
    Args:
      keg - the keg instance to chart
    """
    keg, stats = self._get_keg_stats(context)

    volmap = stats.get('volume_by_day_of_week')
    if not volmap:
      raise ChartUnavailableError, "Daily volumes unavailable"

    vals = []
    for k in sorted(volmap.keys()):
      vals.append(float(volmap[k]))
    if len(vals) != 7:
      raise ChartUnavailableError, "Volume data is corrupt"
    chart = self._day_of_week_chart(vals)
    return chart.get_url()


  def _day_of_week_chart(self, vals):
    chart = pygooglechart.StackedVerticalBarChart(self._width, self._height)
    chart.set_bar_width(20)
    chart.set_colours(['4D89F9','C6D9FD'])

    # convert from 0=Monday to 0=Sunday
    vals.insert(0, vals.pop(-1))

    chart.add_data(vals)
    labels = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']
    chart.set_axis_labels('x', labels)
    return chart

  def chart_sessions_weekday(self, context):
    """ Vertical bar chart showing session volume by day of week.

    Syntax:
      {% chart sessions_weekday <user> width height %}
    Args:
      sessions - an iterable of DrinkingSession or UserDrinkingSessionPart
                 instances
    """
    user_var = Variable(self._args[0])
    user = user_var.resolve(context)
    if not user:
      raise ChartUnavailableError, "Must give a user as argument"
    chunks = user.user_session_chunks.all()

    weekdays = [0] * 7

    for chunk in chunks:
      weekday = chunk.starttime.weekday()
      weekdays[weekday] += float(chunk.Volume())
    chart = self._day_of_week_chart(weekdays)
    return chart.get_url()

  def chart_sessions_volume(self, context):
    """ Line chart showing session volumes.

    Syntax:
      {% chart sessions_volume <sessions> width height %}
    Args:
      sessions - an iterable of DrinkingSession or UserDrinkingSessionPart
                 instances
    """
    sessions_var = Variable(self._args[0])
    sessions = sessions_var.resolve(context)
    if not sessions:
      raise ChartUnavailableError, "Must give sessions as argument"

    data = []
    avgs = []
    avg_points = []
    total = 0.0
    for sess in sessions:
      vol = int(sess.Volume())
      data.append(vol)
      avg_points.append(vol)
      avg_points = avg_points[-5:]
      avgs.append(sum(avg_points) / len(avg_points))

    chart = pygooglechart.SimpleLineChart(self._width, self._height)
    chart.set_colours(['4D89F9','C6D9FD'])
    chart.add_data(data)
    chart.add_data(avgs)
    return chart.get_url()

  def chart_users_by_volume(self, context):
    """Pie chart showing users by volume.

    Syntax:
      {% chart users_by_volume keg width height %}
    """
    chart = pygooglechart.PieChart2D(300, 100)
    keg, stats = self._get_keg_stats(context)
    volmap = stats.get('volume_by_drinker', {})
    if not volmap:
      raise ChartUnavailableError, "no data"

    data = []
    for username, volume_ml in volmap.iteritems():
      data.append((volume_ml, username))

    other_vol = 0
    for volume_ml, username in data[10:]:
      other_vol += volume_ml
    data = data[:10]

    if other_vol:
      data.append((other_vol, 'all others'))

    data.sort(reverse=True)

    labels = []
    data_vals = []
    for volume_ml, username in data:
      ounces = units.Quantity(volume_ml, units=units.UNITS.Milliliter)
      ounces = ounces.ConvertTo.Ounce
      data_vals.append(ounces)
      label = '%s (%ioz)' % (username, ounces)
      labels.append(label)
    chart.add_data(data_vals)
    chart.set_pie_labels(labels)
    return chart.get_url()


class QueryNode(Node):
  def __init__(self, var_name, queryset, limit):
    self._queryset = queryset
    self._var_name = var_name
    self._limit_items = limit

  def render(self, context):
    context[self._var_name] = list(self._queryset[:self._limit_items])
    return ''

@register.filter
def bac_format(text):
  try:
    f = float(text)
  except ValueError:
    return ''
  BAC_MAX = 0.16
  STEPS = 32
  colors = ['#%02x0000' % (x*8,) for x in range(STEPS)]
  bacval = min(max(0, f), BAC_MAX)
  colorstep = BAC_MAX/float(STEPS)
  color = colors[min(STEPS-1, int(bacval/colorstep))]
  ret = '<font color="%s">%.3f</font>' % (color, f)
  if f > 0.08:
    ret = '<b>%s</b>' % ret
  return mark_safe(ret)

@register.filter
def datetime_js(val):
  if not isinstance(val, datetime.datetime):
    return "new Date(1970,0,1)"
  vals = map(str, (
      val.year,
      val.month - 1,
      val.day,
      val.hour,
      val.minute,
      val.second))
  return "new Date(%s)" % (",".join(vals))



# -*- coding: latin-1 -*-
# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import random

from django.conf import settings
from django.core import urlresolvers
from django.db import models
from django.db.models.signals import post_save
from django.db.models.signals import pre_save
from django.contrib.sites.models import Site
from django.contrib.auth.models import User

from pykeg.core import kb_common
from pykeg.core import fields
from pykeg.core import stats
from pykeg.core import units
from pykeg.core import util

from pykeg.beerdb import models as bdb

"""Django models definition for the kegbot database."""

def mugshot_file_name(instance, filename):
  rand_salt = random.randrange(0xffff)
  new_filename = '%04x-%s' % (rand_salt, filename)
  return os.path.join('mugshots', instance.user.username, new_filename)

def _set_seqn_pre_save(sender, instance, **kwargs):
  if instance.seqn:
    return
  prev = sender.objects.filter(site=instance.site).order_by('-seqn')
  if not prev.count():
    seqn = 1
  else:
    seqn = prev[0].seqn + 1
  instance.seqn = seqn


class KegbotSite(models.Model):
  name = models.CharField(max_length=64, unique=True,
      help_text='A short name for this site, eg "default" or "sfo"')
  description = models.TextField(blank=True, null=True,
      help_text='Description of this site')

  def __str__(self):
    return '%s %s' % (self.name, self.description)

class UserPicture(models.Model):
  def __str__(self):
    return "%s UserPicture" % (self.user,)

  user = models.ForeignKey(User)
  image = models.ImageField(upload_to=mugshot_file_name)
  active = models.BooleanField(default=True)


class UserProfile(models.Model):
  """Extra per-User information."""
  GENDER_CHOICES = (
    ('male', 'male'),
    ('female', 'female'),
  )
  def __str__(self):
    return "profile for %s" % (self.user,)

  def FacebookProfile(self):
    if 'socialregistration' not in settings.INSTALLED_APPS:
      return None
    qs = self.user.facebookprofile_set.all()
    if qs:
      return qs[0]

  def TwitterProfile(self):
    if 'socialregistration' not in settings.INSTALLED_APPS:
      return None
    qs = self.user.twitterprofile_set.all()
    if qs:
      return qs[0]

  def MugshotUrl(self):
    if self.mugshot:
      img_url = self.mugshot.image.url
    else:
      args = ('images/unknown-drinker.png',)
      img_url = urlresolvers.reverse('site-media', args=args)
    return img_url

  user = models.OneToOneField(User)
  gender = models.CharField(max_length=8, choices=GENDER_CHOICES)
  weight = models.FloatField()
  mugshot = models.ForeignKey(UserPicture, blank=True, null=True)

def user_post_save(sender, instance, **kwargs):
  defaults = {
    'weight': kb_common.DEFAULT_NEW_USER_WEIGHT,
    'gender': kb_common.DEFAULT_NEW_USER_GENDER,
  }
  profile, new = UserProfile.objects.get_or_create(user=instance,
      defaults=defaults)
post_save.connect(user_post_save, sender=User)


class KegSize(models.Model):
  """ A convenient table of common Keg sizes """
  def Volume(self):
    return units.Quantity(self.volume_ml, units.RECORD_UNIT)

  def __str__(self):
    return "%s [%.2f gal]" % (self.name, self.Volume().ConvertTo.USGallon)

  name = models.CharField(max_length=128)
  volume_ml = models.FloatField()


class KegTap(models.Model):
  """A physical tap of beer."""
  site = models.ForeignKey(KegbotSite)
  seqn = models.PositiveIntegerField()
  name = models.CharField(max_length=128)
  meter_name = models.CharField(max_length=128)
  ml_per_tick = models.FloatField(default=(1000.0/2200.0))
  description = models.TextField(blank=True, null=True)
  current_keg = models.ForeignKey('Keg', blank=True, null=True)
  max_tick_delta = models.PositiveIntegerField(default=100)
  temperature_sensor = models.ForeignKey('ThermoSensor', blank=True, null=True)

  def __str__(self):
    return "%s: %s" % (self.meter_name, self.name)

  def Temperature(self):
    if self.temperature_sensor:
      last_rec = self.temperature_sensor.thermolog_set.all().order_by('-time')
      if last_rec:
        return last_rec[0]
    return None

pre_save.connect(_set_seqn_pre_save, sender=KegTap)

class Keg(models.Model):
  """ Record for each installed Keg. """
  class Meta:
    unique_together = ('site', 'seqn')

  def full_volume(self):
    return self.size.Volume()

  def served_volume(self):
    drinks = Drink.objects.filter(keg__exact=self, status__exact='valid')
    tot = units.Quantity(0, units.RECORD_UNIT)
    for d in drinks:
      tot += d.Volume()
    return tot

  def remaining_volume(self):
    return self.full_volume() - self.served_volume()

  def percent_full(self):
    return float(self.remaining_volume()) / float(self.full_volume()) * 100

  def keg_age(self):
    if self.status == 'online':
      end = datetime.datetime.now()
    else:
      end = self.enddate
    return end - self.startdate

  def is_empty(self):
    return float(self.remaining_volume()) <= 0

  def tap(self):
    q = self.kegtap_set.all()
    if q:
      return q[0]
    return None

  def previous(self):
    q = Keg.objects.filter(startdate__lt=self.startdate).order_by('-startdate')
    if q.count():
      return q[0]
    return None

  def next(self):
    q = Keg.objects.filter(startdate__gt=self.startdate).order_by('startdate')
    if q.count():
      return q[0]
    return None

  def GetStats(self):
    if hasattr(self, '_stats'):
      return self._stats
    qs = self.stats.all()
    if qs:
      self._stats = qs[0].stats
    else:
      self._stats = {}
    return self._stats

  def Sessions(self):
    chunks = SessionChunk.objects.filter(keg=self)
    res = list(set(c.session for c in chunks))
    res.sort()
    return res

  def TopDrinkers(self):
    stats = self.GetStats()
    if not stats:
      return []
    ret = []
    volmap = stats.get('volume_by_drinker', {})
    for username, vol in volmap.iteritems():
      username = str(username)
      vol = float(vol)
      try:
        user = User.objects.get(username=username)
      except User.DoesNotExist:
        continue  # should not happen
      volume = units.Quantity(vol)
      ret.append((volume, user))
    ret.sort(reverse=True)
    return ret

  def __str__(self):
    return "Keg #%s - %s" % (self.id, self.type)

  site = models.ForeignKey(KegbotSite, related_name='kegs')
  seqn = models.PositiveIntegerField()
  type = models.ForeignKey(bdb.BeerType)
  size = models.ForeignKey(KegSize)
  startdate = models.DateTimeField('start date', default=datetime.datetime.now)
  enddate = models.DateTimeField('end date', default=datetime.datetime.now)
  status = models.CharField(max_length=128, choices=(
     ('online', 'online'),
     ('offline', 'offline'),
     ('coming soon', 'coming soon')))
  description = models.CharField(max_length=256, blank=True, null=True)
  origcost = models.FloatField(default=0, blank=True, null=True)
  notes = models.TextField(blank=True, null=True,
      help_text='Private notes about this keg, viewable only by admins.')

def _KegPreSave(sender, instance, **kwargs):
  keg = instance

  # We don't need to do anything if the keg is still online.
  if keg.status != 'offline':
    return

  # Determine first drink date & set keg start date to it if earlier.
  drinks = instance.drinks.all().order_by('endtime')
  if drinks:
    drink = drinks[0]
    if drink.endtime < instance.startdate:
      instance.startdate = drink.endtime

  # Determine last drink date & set keg end date to it if later.
  drinks = instance.drinks.all().order_by('-endtime')
  if drinks:
    drink = drinks[0]
    if drink.endtime > instance.enddate:
      instance.enddate = drink.endtime

pre_save.connect(_set_seqn_pre_save, sender=Keg)
pre_save.connect(_KegPreSave, sender=Keg)


class DrinkManager(models.Manager):
  def valid(self):
    return self.filter(status='valid')

class Drink(models.Model):
  """ Table of drinks records """
  class Meta:
    unique_together = ('site', 'seqn')
    get_latest_by = 'endtime'
    ordering = ('-endtime',)

  def Volume(self):
    return units.Quantity(self.volume_ml, units.RECORD_UNIT)

  def GetSession(self):
    return self.session

  def PourDuration(self):
    return self.endtime - self.starttime

  def ShortUrl(self):
    domain = Site.objects.get_current().domain
    return 'http://%s/d/%i' % (domain, self.id)

  def calories(self):
    if not self.keg or not self.keg.type:
      return 0
    cal = self.keg.type.calories_oz * self.Volume().ConvertTo.Ounce
    return cal

  def bac(self):
    bacs = self.bac_set.all()
    if bacs.count() == 1:
      return bacs[0].bac
    return 0

  def __str__(self):
    return "Drink %s by %s" % (self.id, self.user)

  objects = DrinkManager()

  site = models.ForeignKey(KegbotSite, related_name='drinks')
  seqn = models.PositiveIntegerField()

  # Ticks records the actual meter reading, which is never changed once
  # recorded.
  ticks = models.PositiveIntegerField()

  # Volume is the actual volume of the drink.  Its initial value is a function
  # of `ticks`, but it may be adjusted, eg due to calibration or mis-recording.
  volume_ml = models.FloatField()

  # Similarly, recording both the start and end times of a drink may seem odd.
  # The idea was to someday add metrics to the web page showing pour speeds.
  # This was never terribly exciting so it didn't happen, but space is cheap
  # so I'm inclined to keep the data rather than chuck it.
  #
  # For sorting and other operations requiring a single date, the endtime is
  # used.  TODO(mikey): make sure this is actually the case
  starttime = models.DateTimeField()
  endtime = models.DateTimeField()
  user = models.ForeignKey(User, null=True, blank=True, related_name='drinks')
  keg = models.ForeignKey(Keg, null=True, blank=True, related_name='drinks')
  status = models.CharField(max_length=128, choices = (
     ('valid', 'valid'),
     ('invalid', 'invalid'),
     ), default = 'valid')
  session = models.ForeignKey('DrinkingSession',
      related_name='drinks', null=True, blank=True, editable=False)
  auth_token = models.CharField(max_length=256, blank=True, null=True)
  duration = models.PositiveIntegerField(blank=True, null=True)

  def _UpdateUserStats(self):
    if self.user:
      defaults = {
        'site': self.site,
      }
      stats, created = self.user.stats.get_or_create(defaults=defaults)
      stats.Update(self)

  def _UpdateKegStats(self):
    if self.keg:
      defaults = {
        'site': self.site,
      }
      stats, created = self.keg.stats.get_or_create(defaults=defaults)
      stats.Update(self)

  def PostProcess(self):
    self._UpdateUserStats()
    self._UpdateKegStats()
    SystemEvent.ProcessDrink(self)

pre_save.connect(_set_seqn_pre_save, sender=Drink)

class AuthenticationToken(models.Model):
  """A secret token to authenticate a user, optionally pin-protected."""
  class Meta:
    unique_together = ('site', 'seqn', 'auth_device', 'token_value')

  def __str__(self):
    ret = "%s: %s" % (self.auth_device, self.token_value)
    if self.user is not None:
      ret = "%s (%s)" % (ret, self.user.username)
    return ret

  site = models.ForeignKey(KegbotSite, related_name='tokens')
  seqn = models.PositiveIntegerField()
  auth_device = models.CharField(max_length=64)
  token_value = models.CharField(max_length=128)
  pin = models.CharField(max_length=256, blank=True, null=True)
  user = models.ForeignKey(User, blank=True, null=True)
  created = models.DateTimeField(auto_now_add=True)
  enabled = models.BooleanField(default=True)
  expires = models.DateTimeField(blank=True, null=True)

  def IsAssigned(self):
    return self.user is not None

  def IsActive(self):
    if not self.enabled:
      return False
    if not self.expires:
      return True
    return datetime.datetime.now() < self.expires

pre_save.connect(_set_seqn_pre_save, sender=AuthenticationToken)

class BAC(models.Model):
  """ Calculated table of instantaneous blood alcohol estimations.

  A BAC value may be added by Kegbot after each pour of a registered drinker.
  A relationship to the drink causing the calculation is stored, to ease
  lookup by Drink.

  TODO(mikey): consider making the Drink optional so that intermediate values
  can be added. Seems of dubious utility.
  TODO(mikey): if Drink is mandatory, should eliminated rectime.
  """
  user = models.ForeignKey(User)
  drink = models.ForeignKey(Drink)
  rectime = models.DateTimeField()
  bac = models.FloatField()

  def __str__(self):
    return "%s BAC at %s" % (self.user, self.drink)

  @classmethod
  def ProcessDrink(cls, d):
    """Generate the bac for a drink"""
    bac_qs = d.bac_set.all()
    if len(bac_qs):
      # Already has a recorded BAC.
      return bac_qs[0]

    if not d.user:
      return

    try:
      profile = d.user.get_profile()
    except UserProfile.DoesNotExist:
      # can't compute bac if there is no profile
      return
    if not d.keg:
      return
    if not d.keg.type.abv:
      return

    # TODO: delete/update previous bac for this drink iff exists

    # consider previously recorded bac for non-guest users
    matches = BAC.objects.filter(user=d.user).order_by('-rectime')
    prev_bac = 0
    if len(matches):
      last_bac = matches[0]
      prev_bac = util.decomposeBAC(last_bac.bac, (d.endtime - last_bac.rectime).seconds)

    now = util.instantBAC(profile.gender, profile.weight, d.keg.type.abv,
                          d.Volume().ConvertTo.Ounce)
    b = BAC(user=d.user, drink=d, rectime=d.endtime, bac=now+prev_bac)
    b.save()
    return b


def find_object_in_window(qset, when, window):
  matching = qset.filter(
      starttime__lte=when + window,
      endtime__gte=when - window
  )

  if matching:
    return matching[0]

  # Match objects ending just before the start
  earlier = qset.filter(
      endtime__gte=(when - window),
      starttime__lt=when,
  )

  if earlier:
    return earlier[0]

  # Match objects occuring just after the end
  newer = qset.filter(
      starttime__lte=(when - window),
      endtime__gt=when,
  )

  if newer:
    return newer[0]

  return None


class SessionManager(models.Manager):
  def valid(self):
    return self.filter(volume_ml__gt=kb_common.MIN_SESSION_VOLUME_DISPLAY_ML)

class AbstractChunk(models.Model):
  class Meta:
    abstract = True
    get_latest_by = 'starttime'
    ordering = ('-starttime',)

  starttime = models.DateTimeField()
  endtime = models.DateTimeField()
  volume_ml = models.FloatField(default=0)

  def Volume(self):
    return units.Quantity(self.volume_ml, units.RECORD_UNIT)

  def Duration(self):
    return self.endtime - self.startime

  def AddDrink(self, drink):
    if self.starttime > drink.starttime:
      self.starttime = drink.starttime
    if self.endtime < drink.endtime:
      self.endtime = drink.endtime
    self.volume_ml += drink.volume_ml
    self.save()


class DrinkingSession(AbstractChunk):
  """A collection of contiguous drinks. """
  class Meta:
    unique_together = ('site', 'seqn')
    get_latest_by = 'starttime'
    ordering = ('-starttime',)

  objects = SessionManager()
  site = models.ForeignKey(KegbotSite, related_name='sessions')
  seqn = models.PositiveIntegerField()

  def __str__(self):
    return "Session #%s: %s" % (self.id, self.starttime)

  def count_drinkers(self):
    return self.user_chunks.filter(user__isnull=False).count()

  def count_pints(self):
    return int(round(int(self.Volume().ConvertTo.Pint)))

  def summarize_drinkers(self):
    def fmt(user):
      url = '/drinker/%s/' % (user.username,)
      return '<a href="%s">%s</a>' % (url, user.username)
    chunks = self.user_chunks.all().order_by('-volume_ml')
    users = tuple(c.user for c in chunks)
    names = tuple(fmt(u) for u in users if u)

    if None in users:
      guest_trailer = ' (and possibly others)'
    else:
      guest_trailer = ''

    num = len(names)
    if num == 0:
      return 'no known drinkers'
    elif num == 1:
      ret = names[0]
    elif num == 2:
      ret = '%s and %s' % names
    elif num == 3:
      ret = '%s, %s and %s' % names
    else:
      if guest_trailer:
        return '%s, %s and at least %i others' % (names[0], names[1], num-2)
      else:
        return '%s, %s and %i others' % (names[0], names[1], num-2)

    return '%s%s' % (ret, guest_trailer)

  def GetTitle(self):
    return 'Session %i' % (self.seqn,)

  def Volume(self):
    return units.Quantity(self.volume_ml, units.RECORD_UNIT)

  def Duration(self):
    return self.endtime - self.startime

  def AddDrink(self, drink):
    super(DrinkingSession, self).AddDrink(drink)
    defaults = {'starttime': drink.starttime, 'endtime': drink.endtime}

    # Update or create a SessionChunk.
    chunk, created = SessionChunk.objects.get_or_create(session=self,
        user=drink.user, keg=drink.keg, defaults=defaults)
    chunk.AddDrink(drink)

    # Update or create a UserSessionChunk.
    chunk, created = UserSessionChunk.objects.get_or_create(session=self,
        user=drink.user, defaults=defaults)
    chunk.AddDrink(drink)

    # Update or create a KegSessionChunk.
    chunk, created = KegSessionChunk.objects.get_or_create(session=self,
        keg=drink.keg, defaults=defaults)
    chunk.AddDrink(drink)

  def UserChunksByVolume(self):
    chunks = self.user_chunks.all().order_by('-volume_ml')
    return chunks

  @classmethod
  def AssignSessionForDrink(cls, drink):
    # Return existing session if already assigned.
    if drink.session:
      return drink.session

    # Return last session if one already exists
    q = drink.site.sessions.all()
    window = datetime.timedelta(minutes=kb_common.DRINK_SESSION_TIME_MINUTES)
    session = find_object_in_window(q, drink.endtime, window)
    if session:
      session.AddDrink(drink)
      drink.session = session
      drink.save()
      return session

    # Create a new session
    session = cls(starttime=drink.starttime, endtime=drink.endtime,
        site=drink.site)
    session.save()
    session.AddDrink(drink)
    drink.session = session
    drink.save()
    return session

pre_save.connect(_set_seqn_pre_save, sender=DrinkingSession)


class SessionChunk(AbstractChunk):
  """A specific user and keg contribution to a session."""
  class Meta:
    unique_together = ('session', 'user', 'keg')
    get_latest_by = 'starttime'
    ordering = ('-starttime',)

  session = models.ForeignKey(DrinkingSession, related_name='chunks')
  user = models.ForeignKey(User, related_name='session_chunks', blank=True,
      null=True)
  keg = models.ForeignKey(Keg, related_name='session_chunks', blank=True,
      null=True)


class UserSessionChunk(AbstractChunk):
  """A specific user's contribution to a session (spans all kegs)."""
  class Meta:
    unique_together = ('session', 'user')
    get_latest_by = 'starttime'
    ordering = ('-starttime',)

  session = models.ForeignKey(DrinkingSession, related_name='user_chunks')
  user = models.ForeignKey(User, related_name='user_session_chunks', blank=True,
      null=True)


class KegSessionChunk(AbstractChunk):
  """A specific keg's contribution to a session (spans all users)."""
  class Meta:
    unique_together = ('session', 'keg')
    get_latest_by = 'starttime'
    ordering = ('-starttime',)

  objects = SessionManager()
  session = models.ForeignKey(DrinkingSession, related_name='keg_chunks')
  keg = models.ForeignKey(Keg, related_name='keg_session_chunks', blank=True,
      null=True)


class ThermoSensor(models.Model):
  class Meta:
    unique_together = ('site', 'seqn')

  site = models.ForeignKey(KegbotSite, related_name='thermosensors')
  seqn = models.PositiveIntegerField()
  raw_name = models.CharField(max_length=256)
  nice_name = models.CharField(max_length=128)

  def __str__(self):
    return self.nice_name

  def LastLog(self):
    try:
      return self.thermolog_set.latest()
    except Thermolog.DoesNotExist:
      return None

pre_save.connect(_set_seqn_pre_save, sender=ThermoSensor)


class Thermolog(models.Model):
  """ A log from an ITemperatureSensor device of periodic measurements. """
  class Meta:
    unique_together = ('site', 'seqn')
    get_latest_by = 'time'

  site = models.ForeignKey(KegbotSite, related_name='thermologs')
  seqn = models.PositiveIntegerField()
  sensor = models.ForeignKey(ThermoSensor)
  temp = models.FloatField()
  time = models.DateTimeField()

  def __str__(self):
    return '%s %.2f C / %.2f F [%s]' % (self.sensor, self.TempC(),
        self.TempF(), self.time)

  def TempC(self):
    return self.temp

  def TempF(self):
    return util.CtoF(self.temp)

  @classmethod
  def CompressLogs(cls):
    now = datetime.datetime.now()

    # keep at least the most recent 24 hours
    keep_time = now - datetime.timedelta(hours=24)

    # round down to the start of the day
    keep_date = datetime.datetime(year=keep_time.year, month=keep_time.month,
        day=keep_time.day)

    print "Compressing thermo logs prior to", keep_date

    def _LogGroup(sensor, records):
      num_readings = len(records)
      temps = [x.temp for x in records]
      min_temp = min(temps)
      max_temp = max(temps)
      sum_temps = sum(temps)
      mean_temp = sum_temps / num_readings

      base_date = records[0].time
      daily_date = datetime.datetime(year=base_date.year,
          month=base_date.month, day=base_date.day)

      print "%s\t%s\t%s\t%s\t%s" % (daily_date, num_readings, min_temp, max_temp,
          mean_temp)
      try:
        ThermoSummaryLog.objects.get(date=daily_date, period='daily')
        # Oops! Group already exists
        print "Warning: log already exists for", daily_date
        return
      except ThermoSummaryLog.DoesNotExist:
        pass
      new_rec = ThermoSummaryLog.objects.create(
          sensor=sensor,
          date=daily_date,
          period='daily',
          num_readings=num_readings,
          min_temp=min_temp,
          max_temp=max_temp,
          mean_temp=mean_temp)

    for sensor in ThermoSensor.objects.all():
      old_entries = sensor.thermolog_set.filter(time__lt=keep_date).order_by('time')

      group = []
      group_date = None
      for entry in old_entries:
        entry_date = datetime.datetime(year=entry.time.year,
            month=entry.time.month, day=entry.time.day)
        if group_date is None:
          group_date = entry_date
        if group_date == entry_date:
          group.append(entry)
        else:
          _LogGroup(sensor, group)
          group = [entry]
          group_date = entry_date

      if group:
        _LogGroup(sensor, group)

      old_entries.delete()

def thermolog_post_save(sender, instance, **kwargs):
  # TODO(mikey): prevent thermolog entries from being changed; if a particular
  # record is saved more than once, the average temperature will be invalid.
  daily_date = datetime.datetime(year=instance.time.year,
      month=instance.time.month,
      day=instance.time.day)
  defaults = {
      'site': instance.site,
      'num_readings': 1,
      'min_temp': instance.temp,
      'max_temp': instance.temp,
      'mean_temp': instance.temp,
  }
  daily_log, created = ThermoSummaryLog.objects.get_or_create(
      sensor=instance.sensor,
      period='daily',
      date=daily_date,
      defaults=defaults)

  if not created:
    new_mean = daily_log.num_readings * daily_log.mean_temp + instance.temp
    new_mean /= daily_log.num_readings + 1
    daily_log.mean_temp = new_mean

    daily_log.num_readings += 1

    if instance.temp > daily_log.max_temp:
      daily_log.max_temp = instance.temp
    if instance.temp < daily_log.min_temp:
      daily_log.min_temp = instance.temp

  daily_log.save()

pre_save.connect(_set_seqn_pre_save, sender=Thermolog)
post_save.connect(thermolog_post_save, sender=Thermolog)


class ThermoSummaryLog(models.Model):
  """A summarized temperature sensor log."""
  class Meta:
    unique_together = ('site', 'seqn')

  PERIOD_CHOICES = (
    ('daily', 'daily'),
  )
  site = models.ForeignKey(KegbotSite, related_name='thermosummarylogs')
  seqn = models.PositiveIntegerField()
  sensor = models.ForeignKey(ThermoSensor)
  date = models.DateTimeField()
  period = models.CharField(max_length=64, choices=PERIOD_CHOICES,
      default='daily')
  num_readings = models.PositiveIntegerField()
  min_temp = models.FloatField()
  max_temp = models.FloatField()
  mean_temp = models.FloatField()

pre_save.connect(_set_seqn_pre_save, sender=ThermoSummaryLog)


class RelayLog(models.Model):
  """ A log from an IRelay device of relay events/ """
  class Meta:
    unique_together = ('site', 'seqn')

  site = models.ForeignKey(KegbotSite, related_name='relaylogs')
  seqn = models.PositiveIntegerField()
  name = models.CharField(max_length=128)
  status = models.CharField(max_length=32)
  time = models.DateTimeField()

pre_save.connect(_set_seqn_pre_save, sender=RelayLog)


class Config(models.Model):
  def __str__(self):
    return '%s=%s' % (self.key, self.value)

  site = models.ForeignKey(KegbotSite, related_name='configs')
  key = models.CharField(max_length=255, unique=True)
  value = models.TextField()

  @classmethod
  def get(cls, key, default=None):
    try:
      return cls.objects.get(key=key)
    except cls.DoesNotExist:
      return default


class _StatsModel(models.Model):
  STATS_BUILDER = None
  class Meta:
    abstract = True
  site = models.ForeignKey(KegbotSite)
  date = models.DateTimeField(default=datetime.datetime.now)
  stats = fields.JSONField()
  revision = models.PositiveIntegerField(default=0)

  def Update(self, obj):
    builder = self.STATS_BUILDER()
    # TODO(mikey): use prev
    self.stats = builder.Build(obj)
    self.revision = builder.REVISION
    self.save()


class UserStats(_StatsModel):
  STATS_BUILDER = stats.DrinkerStatsBuilder
  user = models.ForeignKey(User, unique=True, related_name='stats')

  def __str__(self):
    return 'UserStats for %s' % self.user


class KegStats(_StatsModel):
  STATS_BUILDER = stats.KegStatsBuilder
  keg = models.ForeignKey(Keg, unique=True, related_name='stats')

  def __str__(self):
    return 'KegStats for %s' % self.keg


class SystemEvent(models.Model):
  class Meta:
    ordering = ('-when', '-id')
    get_latest_by = 'when'

  KINDS = (
      ('drink_poured', 'Drink poured'),
      ('session_started', 'Session started'),
      ('session_joined', 'User joined session'),
      ('keg_tapped', 'Keg tapped'),
      ('keg_ended', 'Keg ended'),
  )

  site = models.ForeignKey(KegbotSite)
  seqn = models.PositiveIntegerField()
  kind = models.CharField(max_length=255, choices=KINDS,
      help_text='Type of event.')
  when = models.DateTimeField(help_text='Time of the event.')
  user = models.ForeignKey(User, blank=True, null=True,
      related_name='events',
      help_text='User responsible for the event, if any.')
  drink = models.ForeignKey(Drink, blank=True, null=True,
      related_name='events',
      help_text='Drink involved in the event, if any.')
  keg = models.ForeignKey(Keg, blank=True, null=True,
      related_name='events',
      help_text='Keg involved in the event, if any.')
  session = models.ForeignKey(DrinkingSession, blank=True, null=True,
      related_name='events',
      help_text='Session involved in the event, if any.')

  @classmethod
  def ProcessDrink(cls, drink):
    keg = drink.keg
    site = drink.site
    if keg:
      q = keg.events.filter(kind='keg_started')
      if q.count() == 0:
        e = keg.events.create(site=site, kind='keg_started', when=drink.endtime,
            keg=keg)
        e.save()

    session = drink.session
    if session:
      q = session.events.filter(kind='session_started')
      if q.count() == 0:
        e = session.events.create(site=site, kind='session_started',
            when=session.starttime, user=drink.user)
        e.save()

    user = drink.user
    if user:
      q = user.events.filter(kind='session_joined')
      if q.count() == 0:
        e = user.events.create(site=site, kind='session_joined',
            when=drink.endtime, user=user)
        e.save()

    q = drink.events.filter(kind='drink_poured')
    if q.count() == 0:
      e = drink.events.create(site=site, kind='drink_poured',
          when=drink.endtime, drink=drink, user=drink.user, keg=drink.keg,
          session=drink.session)
      e.save()

pre_save.connect(_set_seqn_pre_save, sender=SystemEvent)



#!/usr/bin/env python

from distribute_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages

VERSION = "0.7.5"
SHORT_DESCRIPTION = "Kegbot kegerator controller software"
LONG_DESCRIPTION = """This package contains Kegbot core controller and Django
frontend package.

Kegbot is a hardware and software system to record and monitor access to a beer
kegerator.  For more information and documentation, see http://kegbot.org/

**Note:** This package is still *experimental* and subject to change.
"""

setup(
    name = "kegbot",
    version = VERSION,
    description = SHORT_DESCRIPTION,
    long_description = LONG_DESCRIPTION,
    author = "mike wakerly",
    author_email = "opensource@hoho.com",
    url = "http://kegbot.org/",
    packages = find_packages('src'),
    package_dir = {
      '' : 'src',
    },
    scripts = [
      'distribute_setup.py',
      'src/pykeg/bin/fb_publisher.py',
      'src/pykeg/bin/kegboard_daemon.py',
      'src/pykeg/bin/kegboard_monitor.py',
      'src/pykeg/bin/kegbot_admin.py',
      'src/pykeg/bin/kegbot_core.py',
      'src/pykeg/bin/kegbot_master.py',
      'src/pykeg/bin/kegbot_twitter.py',
      'src/pykeg/bin/lcd_daemon.py',
      'src/pykeg/bin/rfid_daemon.py',
      'src/pykeg/bin/sound_server.py',
    ],
    install_requires = [
      'django >= 1.2',
      'django-imagekit >= 0.3.3',
      'django-registration',
      'django_extensions',

      # NOTE(mikey): socialregistration does not yet declare the
      # facebook-python-sdk prerequisite for itself.
      'facebook-python-sdk',
      'django-socialregistration >= 0.4.2',

      #'MySQL-python',
      #'pil',
      'protobuf >= 2.3.0-1',
      'pylcdui >= 0.5.5',
      #'pysqlite>=2.0.3',
      'python-gflags >= 1.3',
      'South >= 0.7',
      'Sphinx',
      #'django_nose',
      #'python-openid >= 2.2.5',  # removeme once PIL package works
      'pygooglechart',
    ],
    dependency_links = [
        'http://dist.repoze.org/PIL-1.1.6.tar.gz',
        'http://kegbot.org/kmedia/protobuf-2.3.0-1.tgz',
        'http://kegbot.org/kmedia/python-openid-2.2.5.tgz',

        # Self-maintained package due to upstream issue:
        # http://github.com/facebook/python-sdk/issues/#issue/21
        'http://github.com/downloads/mik3y/python-sdk/facebook-python-sdk-0.1.tar.gz',
    ],
    include_package_data = True,

)



from pykeg.core import features

def enabled_features(request):
  """Adds a USE_FEATURENAME flags for each enabled feature.

  Currently defined:
    USE_FACEBOOK - Allow registration and fbconnect login via django-socialauth
    USE_TWITTER - pykeg.contrib.twitter is installed.
  """
  # TODO(mikey): this might make it harder to diagnose why features aren't
  # visible/being used.
  ret = {}
  ret['USE_FACEBOOK'] = features.use_facebook()
  ret['USE_TWITTER'] = features.use_twitter()
  return ret

def kbsite(request):
  ret = {}
  ret['kbsite'] = request.kbsite
  return ret



#!/usr/bin/env python

"""Unittest for kegbot module"""

import commands
import datetime
import time
import logging
import socket
import unittest
import kegbot

from django.test import TestCase

from pykeg.core import defaults
from pykeg.core import event
from pykeg.core import models
from pykeg.core import kb_common
from pykeg.core import units
from pykeg.core.net import kegnet
from pykeg.core.net import kegnet_pb2

from pykeg.beerdb import models as bdb_models

LOGGER = logging.getLogger('unittest')

class KegbotTestCase(TestCase):
  def setUp(self):
    del logging.root.handlers[:]
    defaults.set_defaults()
    #defaults.gentestdata()

    models.Drink.objects.all().delete()

    self.keg_size = models.KegSize.objects.create(name='10Liter Keg',
        volume_ml=units.Quantity(10000))

    self.beer_style = bdb_models.BeerStyle.objects.create(name='test style')
    self.brewer = bdb_models.Brewer.objects.create(name='test brewer')
    self.beer_type = bdb_models.BeerType.objects.create(name='test beer type',
        brewer=self.brewer, style=self.beer_style, calories_oz=10.0,
        carbs_oz=2.0, abv=5.0)

    self.test_keg = models.Keg.objects.create(type=self.beer_type,
        size=self.keg_size, status='online', description='None')

    self.test_tap = models.KegTap.objects.create(name='Test Tap',
        meter_name='test_meter_name', ml_per_tick=(1000.0/2200.0),
        current_keg=self.test_keg)

    self.kegbot = kegbot.KegbotCoreApp()
    self.env = self.kegbot._env
    self.backend = self.env.GetBackend()

    self.test_user = self.backend.CreateNewUser('tester')
    self.test_token = self.backend.CreateAuthToken('core.onewire',
        '0000111122223333', 'tester')

    self.test_user_2 = self.backend.CreateNewUser('tester_2')
    self.test_token_2 = self.backend.CreateAuthToken('core.onewire',
        'baadcafebeeff00d', 'tester_2')
    kb_common.AUTH_TOKEN_MAX_IDLE['core.onewire'] = 2

    # Kill the kegbot flow processing thread so we can single step it.
    self.service_thread = self.env._service_thread
    self.env._threads.remove(self.service_thread)

    self.kegbot._Setup()
    self.kegbot._StartThreads()

    self.client = kegnet.KegnetClient()
    while True:
      if self.client.Reconnect():
        break

  def tearDown(self):
    for thr in self.env.GetThreads():
      self.assert_(thr.isAlive(), "thread %s died unexpectedly" % thr.getName())
    self.kegbot.Quit()
    for thr in self.env.GetThreads():
      self.assert_(not thr.isAlive(), "thread %s stuck" % thr.getName())
    del self.kegbot
    del self.env

  def testSimpleFlow(self):
    # Synthesize a 2200-tick flow. The FlowManager should zero on the initial
    # reading of 1000.
    meter_name = self.test_tap.meter_name
    self.client.SendFlowStart(meter_name)
    self.client.SendMeterUpdate(meter_name, 1000)
    count = 0
    while count < 2200:
      count += 100
      self.client.SendMeterUpdate(meter_name, 1000+count)

    self.client.SendFlowStop(meter_name)
    self.service_thread._FlushEvents()

    drinks = self.test_keg.drinks.all()
    self.assertEquals(len(drinks), 1)
    last_drink = drinks[0]

    LOGGER.info('last drink: %s' % (last_drink,))
    self.assertEquals(last_drink.ticks, 2200)

    self.assertEquals(last_drink.keg, self.test_keg)

    self.assertEquals(last_drink.user, None)

  def testAuthenticatedFlow(self):
    meter_name = self.test_tap.meter_name

    # Start a flow by pouring a few ticks.
    self.client.SendMeterUpdate(meter_name, 100)
    self.client.SendMeterUpdate(meter_name, 200)
    time.sleep(1.0)
    self.service_thread._FlushEvents()

    # The flow should be anonymous, since no user is authenticated.
    flow_mgr = self.env.GetFlowManager()
    flows = flow_mgr.GetActiveFlows()
    self.assertEquals(len(flows), 1)
    flow = flows[0]
    self.assertEquals(flow.GetUser(), None)

    # Now authenticate the user.
    # TODO(mikey): should use tap name rather than meter name.
    self.client.SendAuthTokenAdd(self.test_tap.meter_name,
        auth_device_name=self.test_token.auth_device,
        token_value=self.test_token.token_value)
    time.sleep(1.0) # TODO(mikey): need a synchronous wait
    self.service_thread._FlushEvents()
    self.assertEquals(flow.GetUser(), self.test_user)

    # If another user comes along, he takes over the flow.
    self.client.SendAuthTokenAdd(self.test_tap.meter_name,
        auth_device_name=self.test_token_2.auth_device,
        token_value=self.test_token_2.token_value)
    time.sleep(1.0) # TODO(mikey): need a synchronous wait
    self.service_thread._FlushEvents()

    flows = flow_mgr.GetActiveFlows()
    self.assertEquals(len(flows), 1)
    flow = flows[0]
    self.assertEquals(flow.GetUser(), self.test_user_2)

    self.client.SendMeterUpdate(meter_name, 300)
    time.sleep(1.0) # TODO(mikey): need a synchronous wait
    self.service_thread._FlushEvents()
    self.client.SendFlowStop(meter_name)
    time.sleep(1.0) # TODO(mikey): need a synchronous wait
    self.service_thread._FlushEvents()

    drinks = self.test_keg.drinks.all().order_by('id')
    self.assertEquals(len(drinks), 2)
    self.assertEquals(protolib.ToProto(drinks[0].user), self.test_user)
    self.assertEquals(protolib.ToProto(drinks[1].user), self.test_user_2)

  def testIdleFlow(self):
    meter_name = self.test_tap.meter_name
    self.client.SendFlowStart(meter_name)
    self.client.SendMeterUpdate(meter_name, 0)
    self.client.SendMeterUpdate(meter_name, 100)
    time.sleep(1.0)
    self.service_thread._FlushEvents()

    flows = self.env.GetFlowManager().GetActiveFlows()
    self.assertEquals(len(flows), 1)

    # Rewind the flow activity clocks to simulate idleness.
    flows[0]._start_time -= datetime.timedelta(minutes=10)
    flows[0]._end_time = flows[0]._start_time

    # Wait for the heartbeat event to kick in.
    time.sleep(1.0)
    self.service_thread._FlushEvents()

    # Verify that the flow has been ended.
    flows = self.env.GetFlowManager().GetActiveFlows()
    self.assertEquals(len(flows), 0)

  def testTokenDebounce(self):
    meter_name = self.test_tap.meter_name
    self.client.SendFlowStart(meter_name)
    self.client.SendMeterUpdate(meter_name, 0)
    self.client.SendMeterUpdate(meter_name, 100)
    time.sleep(1.0)
    self.service_thread._FlushEvents()

    self.client.SendAuthTokenAdd(self.test_tap.meter_name,
        auth_device_name=self.test_token.auth_device,
        token_value=self.test_token.token_value)
    time.sleep(1.0) # TODO(mikey): need a synchronous wait
    self.service_thread._FlushEvents()

    flows = self.env.GetFlowManager().GetActiveFlows()
    self.assertEquals(len(flows), 1)
    flow = flows[0]
    self.assertEquals(flow.GetUser(), self.test_user)
    original_flow_id = flow.GetId()

    LOGGER.info('Removing token...')
    self.client.SendAuthTokenRemove(self.test_tap.meter_name,
        auth_device_name=self.test_token.auth_device,
        token_value=self.test_token.token_value)
    time.sleep(0.5)
    self.service_thread._FlushEvents()

    # The flow should still be active.
    flows = self.env.GetFlowManager().GetActiveFlows()
    self.assertEquals(len(flows), 1)
    flow = flows[0]
    self.assertEquals(flow.GetId(), original_flow_id)
    self.assertEquals(flow.GetState(), kegnet_pb2.FlowUpdate.ACTIVE)

    # Re-add the token; should be unchanged.
    self.client.SendAuthTokenAdd(self.test_tap.meter_name,
        auth_device_name=self.test_token.auth_device,
        token_value=self.test_token.token_value)
    time.sleep(0.5)
    self.service_thread._FlushEvents()

    # No change in flow
    flows = self.env.GetFlowManager().GetActiveFlows()
    self.assertEquals(len(flows), 1)
    flow = flows[0]
    self.assertEquals(flow.GetId(), original_flow_id)
    self.assertEquals(flow.GetState(), kegnet_pb2.FlowUpdate.ACTIVE)

    # Idle out. TODO(mikey): shift clock instead of sleeping.
    time.sleep(2.5)
    self.service_thread._FlushEvents()
    flows = self.env.GetFlowManager().GetActiveFlows()
    self.assertEquals(len(flows), 0)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
#
# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""SoundServer application implementation."""

import os
import ossaudiodev
import random
import Queue
import urllib2

import gflags
import mad

from pykeg.core import kb_app
from pykeg.core import units
from pykeg.core import util
from pykeg.core.net import kegnet

from pykeg.contrib.soundserver import models

FLAGS = gflags.FLAGS

# TODO(mikey): move this into the db.
gflags.DEFINE_string('base_url', '',
    'Base URL of the main server. If specified, files will be fetched '
    'from this URL, rather than read locally. Experimental.')

gflags.DEFINE_string('cache_dir', '/tmp/sound_server_cache',
    'Directory for cached sound files.')

OUNCE_TRIGGER_THRESHOLDS = (32, 24, 16, 12, 8, 1, 0)

CREDIT_TRIGGER_THRESHOLDS = (100, 20, 10, 5, 1)


class SoundServerApp(kb_app.App):
  def __init__(self, name='sound_server'):
    kb_app.App.__init__(self, name)

  def _Setup(self):
    kb_app.App._Setup(self)

    sound_thread = SoundThread('sound-playback')
    self._AddAppThread(sound_thread)

    self._client = SoundClient(sound_thread)
    self._client_thr = kegnet.KegnetClientThread('kegnet', self._client)
    self._AddAppThread(self._client_thr)


class SoundThread(util.KegbotThread):
  def __init__(self, name):
    util.KegbotThread.__init__(self, name)
    self._file_queue = Queue.Queue()
    self._logger.info('created sound thread')

  def Enqueue(self, filename):
    self._logger.info('Enqueue: %s' % filename)
    self._file_queue.put(filename)

  def ThreadMain(self):
    self._logger.info('in main loop')
    while not self._quit:
      try:
        filename = self._file_queue.get(timeout=0.5)
      except Queue.Empty:
        continue
      self._PlayFile(filename)

  def _PlayFile(self, filename):
    self._logger.info('Playing %s' % filename)
    mf = mad.MadFile(filename)
    dev = ossaudiodev.open('w')
    dev.speed(mf.samplerate())
    dev.setfmt(ossaudiodev.AFMT_S16_LE)
    dev.channels(2)
    while True:
      if self._file_queue.qsize():
        break
      buf = mf.read()
      if buf is None:
        break
      dev.write(buf)


class SoundClient(kegnet.SimpleKegnetClient):
  def __init__(self, sound_thread, addr=None):
    kegnet.SimpleKegnetClient.__init__(self, addr)
    self._sound_thread = sound_thread
    self._event_map = {}
    self._sound_file_map = {}
    self._tempdir = FLAGS.cache_dir
    self._flows = {}

  def PlaySoundFile(self, soundfile):
    local_sound = self._sound_file_map.get(soundfile)
    if not local_sound:
      self._logger.warning('No sound for "%s"' % soundfile)
      return

    self._sound_thread.Enqueue(local_sound)

  def PlaySoundEvent(self, event):
    soundfile = event.soundfile
    self.PlaySoundFile(soundfile)

  def LoadEvents(self):
    all_events = models.SoundEvent.objects.all()
    for event in all_events:
      self.LoadSoundEvent(event)

  def LoadSoundEvent(self, event):
    self._CacheSound(event.soundfile)
    if event.event_name not in self._event_map:
      self._event_map[event.event_name] = list()
    self._event_map[event.event_name].append(event)

  def _CacheSound(self, soundobj):
    if soundobj in self._sound_file_map:
      return

    outfile = os.path.join(self._tempdir, os.path.basename(soundobj.sound.url))
    self._sound_file_map[soundobj] = outfile

    if os.path.exists(outfile):
      self._logger.info('Cached file for "%s" exists at %s' % (soundobj.title, outfile))
      return

    dirname = os.path.dirname(outfile)
    if not os.path.exists(dirname):
      self._logger.info('Creating directory: %s' % dirname)
      os.makedirs(dirname)

    self._logger.info('Saving file "%s" to %s' % (soundobj.title, outfile))
    if FLAGS.base_url:
      url = FLAGS.base_url + str(soundobj.sound.url)
      filedata = urllib2.urlopen(url).read()
    else:
      filedata = soundobj.sound.read()
    outfd = open(outfile, 'wb')
    outfd.write(filedata)
    outfd.close()

  def onCreditAdded(self, event):
    amount = event.amount

    for threshold in CREDIT_TRIGGER_THRESHOLDS:
      if amount < threshold:
        continue
      events = self._GetMatchingEvents('credit.added.threshold', threshold)
      if len(events):
        soundevent = random.choice(events)
        self.PlaySoundFile(soundevent.soundfile)
        return

  def onFlowUpdate(self, event):
    flow_id = event.flow_id

    if event.state == event.COMPLETED:
      if flow_id in self._flows:
        del self._flows[flow_id]
        return

    curr_ounces = float(units.Quantity(event.volume_ml,
        units=units.UNITS.Ounce, from_units=units.UNITS.Milliliter))

    if flow_id not in self._flows:
      self._flows[flow_id] = -1

    last_ounces = self._flows[flow_id]

    for trigger in OUNCE_TRIGGER_THRESHOLDS:
      if (last_ounces < trigger) and (curr_ounces >= trigger):
        self._ThresholdEvent(trigger, event)
        break
    self._flows[flow_id] = curr_ounces

  def _GetMatchingEvents(self, event_name, predicate=None, user=None):
    events = self._event_map.get(event_name, [])
    if predicate is not None:
      events = [e for e in events if e.event_predicate == str(predicate)]
    if user:
      events = [e for e in events if e.user == user]
    return events

  def _ThresholdEvent(self, trigger_ounces, flow_update):
    self._logger.info('Threshold [%i ounces] reached for flow %i' %
        (trigger_ounces, flow_update.flow_id))

    events = self._GetMatchingEvents('flow.threshold.ounces', trigger_ounces)
    self._logger.info('Events matching trigger: %s' % str(events))

    if len(events) == 0:
      return
    elif len(events) == 1:
      soundevent = events[0]
    else:
      soundevent = random.choice(events)

    self.PlaySoundFile(soundevent.soundfile)



# Cloned from django-extensions project: django_extensions.db.fields.json
# django_extensions has the following license:
#
#   Copyright (c) 2007 Michael Trier
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to
#   deal in the Software without restriction, including without limitation the
#   rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#   sell copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#   IN THE SOFTWARE.
#
# Local differences from upstream version:
#   - disabled assert in JSONEncoder
#   - added south introspection rules

"""
JSONField automatically serializes most Python terms to JSON data.
Creates a TEXT field with a default value of "{}".  See test_json.py for
more information.

 from django.db import models
 from django_extensions.db.fields import json

 class LOL(models.Model):
     extra = json.JSONField()
"""

import datetime
from decimal import Decimal
from django.db import models
from django.conf import settings
from django.utils import simplejson
from django.utils.encoding import smart_unicode

class JSONEncoder(simplejson.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Decimal):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            #assert settings.TIME_ZONE == 'UTC'
            return obj.strftime('%Y-%m-%dT%H:%M:%SZ')
        return simplejson.JSONEncoder.default(self, obj)

def dumps(value):
    assert isinstance(value, dict)
    return JSONEncoder().encode(value)

def loads(txt):
    value = simplejson.loads(
        txt,
        parse_float = Decimal,
        encoding    = settings.DEFAULT_CHARSET)
    assert isinstance(value, dict)
    return value

class JSONDict(dict):
    """
    Hack so repr() called by dumpdata will output JSON instead of
    Python formatted data.  This way fixtures will work!
    """
    def __repr__(self):
        return dumps(self)

class JSONField(models.TextField):
    """JSONField is a generic textfield that neatly serializes/unserializes
    JSON objects seamlessly.  Main thingy must be a dict object."""

    # Used so to_python() is called
    __metaclass__ = models.SubfieldBase

    def __init__(self, *args, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = '{}'
        models.TextField.__init__(self, *args, **kwargs)

    def to_python(self, value):
        """Convert our string value to JSON after we load it from the DB"""
        if not value:
            return {}
        elif isinstance(value, basestring):
            res = loads(value)
            assert isinstance(res, dict)
            return JSONDict(**res)
        else:
            return value

    def get_db_prep_save(self, value):
        """Convert our JSON object to a string before we save"""
        if not value:
            return super(JSONField, self).get_db_prep_save("")
        else:
            return super(JSONField, self).get_db_prep_save(dumps(value))

try:
    from south.modelsinspector import add_introspection_rules
except ImportError:
    pass
else:
    add_introspection_rules([
        (
            [JSONField],
            [],
            {},
        ),
    ], ["^pykeg\.core\.jsonfield\.JSONField"])



#!/usr/bin/env python
#
# Copyright 2009 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Master Kegbot process runner.

This program simplifies running multiple Kegbot processes as daemons.
"""

from pykeg.core import importhacks

import gflags

import ConfigParser
import commands
from cStringIO import StringIO
import os
import signal
import sys
import time

from pykeg.core import kb_app
from pykeg.core import util

FLAGS = gflags.FLAGS

gflags.DEFINE_string('master_config', '/etc/kegbot/master.cfg',
    'Path to the master config file.')

gflags.DEFINE_string('logfile_dir', '/var/log/kegbot',
    'Default directory for log files.  This value is used as the '
    'default for internal variable `_logfile_dir`.  An application may '
    'override the default by specifying `logfile = /path/to/file`.')

gflags.DEFINE_string('pidfile_dir', '/var/run/kegbot',
    'Default directory for pid files.  This value is used as the '
    'default for internal variable `_pidfile_dir`.  An application may '
    'override the default by specifying `pidfile = /path/to/file`.')

STATUS_UNKNOWN = -1
STATUS_ALIVE = 1
STATUS_DEAD = 2
STATUS_STOPPED = 3

DEFAULT_CONFIG_DATA = """
[DEFAULT]

### Internal variables

# Default base directory for pid files. Be sure the directory exists and is
# writable by the user running kegbot_master.
_pidfile_dir = %(_flag_pidfile_dir)s

# Default base directory for log files. Be sure the directory exists and is
# writable by the user running kegbot_master.
_logfile_dir = %(_flag_logfile_dir)s

# Default value for whether or not the app is enabled. If the app is not
# enabled, it will not be started, stopped, or show up in status.
_enabled = False

# Name of the program to run. By default, it is inferred from the section name;
# for example, a section named "[kegboard_daemon]" will have _program_name set to
# "kegboard_daemon.py".  You need to override this if you wish to run multiple
# instances of the same program (see 'second_kegboard' example section).
_program_name = %(__name__)s.py

### Default flags

# Location of the pidfile. You probably don't need to change this, provided the
# internal variable _pidfile_dir is something reasonable.
pidfile = %(_pidfile_dir)s/%(__name__)s.pid

# Run the app as a daemon. It would be quite odd to use any value other than
# `True` here...
daemon = True

# Disable stdout logging, since it would not do anything.
log_to_stdout = False

# Enable file logging. You might want to disable this, since there's currently
# no limit on log file sizes (and no built-in log rotation).
log_to_file = True

# Log file path. The default is probably fine, provided the internal variable
# _logfile_dir is set appropriately.
logfile = %(_logfile_dir)s/%(__name__)s.log
"""

def _ExtendPath():
  """Appends directory name of this script to $PATH."""
  local_dir = os.path.dirname(sys.modules[__name__].__file__)
  current_paths = os.environ['PATH'].split(':')
  if local_dir not in current_paths:
    current_paths.append(local_dir)
    os.environ['PATH'] = ':'.join(current_paths)

def _GetConfigParser():
  default_values = {
    '_flag_logfile_dir': FLAGS.logfile_dir,
    '_flag_pidfile_dir': FLAGS.pidfile_dir,
  }
  config = ConfigParser.SafeConfigParser(defaults=default_values)
  default_cfg_fp = StringIO(DEFAULT_CONFIG_DATA)
  config.readfp(default_cfg_fp)
  config.read(FLAGS.master_config)
  return config

class KegbotMasterApp(kb_app.App):
  def _Setup(self):
    kb_app.App._Setup(self)
    _ExtendPath()
    self._config = _GetConfigParser()
    self._commands = {}

    for app in self._EnabledApps():
      self._commands[app] = self._BuildCommand(app)

  def _Enabled(self, app):
    return self._config.getboolean(app, '_enabled')

  def _AllAppNames(self):
    return self._config.sections()

  def _EnabledApps(self):
    apps = self._AllAppNames()
    return (app for app in apps if self._Enabled(app))

  def _BuildCommand(self, app):
    cmd = [self._config.get(app, '_program_name')]
    flags = []

    options = self._config.items(app)
    for flag, value in options:
      if flag.startswith('_'):
        continue
      if value:
        flags.append('--%s=%s' % (flag, value))
      else:
        flags.append('--%s' % (flag,))

    flags.sort()
    ret = cmd + flags
    return ret

  def _PrettyCommand(self, cmdlist):
    options = ''
    command = cmdlist[0]
    if len(cmdlist) > 1:
      options = ' \\\n'.join('    %s' % cmd for cmd in cmdlist[1:])
      return '%s \\\n%s' % (command, options)
    else:
      return command

  def _Usage(self):
    sys.stderr.write('Usage: %s <start|stop|status|configtest>\n' %
        self._extra_argv[0])

  def _StartApps(self):
    self._logger.info('Starting apps')
    for app in self._AllAppNames():
      if app not in self._commands:
        continue
      command = self._commands[app]
      command_str = ' '.join(command)
      self._logger.info('Starting: %s' % app)
      self._logger.info('Command: %s' % command_str)
      status, output = commands.getstatusoutput(command_str)
      if status == 0:
        self._logger.info('Started %s OK' % app)
      else:
        self._logger.error('Error starting %s: %i' % (app, status))
        output_lines = output.splitlines()
        if len(output_lines):
          self._logger.error('Output was:')
          for line in output_lines:
            self._logger.error('<%s>  %s' % (app, line))
        break

  def _AppStatus(self, app):
    pidfile = self._config.get(app, 'pidfile')
    if not os.path.exists(pidfile):
      return STATUS_STOPPED, -1
    try:
      pid_fd = open(pidfile, 'r')
      pid = int(pid_fd.readline().strip())
      alive = util.PidIsAlive(pid)
      if alive:
        return STATUS_ALIVE, pid
      else:
        return STATUS_DEAD, pid
    except (IOError):
      pass
    return STATUS_UNKNOWN, -1

  def _Status(self):
    stats = {}
    for app in self._AllAppNames():
      if app not in self._commands:
        continue
      status, pid = self._AppStatus(app)
      if status == STATUS_ALIVE:
        stats[app] = 'running (%i)' % (pid,)
      elif status == STATUS_DEAD:
        stats[app] = 'died (%i)' % (pid,)
      elif status == STATUS_STOPPED:
        stats[app] = 'stopped'
      else:
        stats[app] = 'unknown'

    print 'Status:'
    for app, status in stats.iteritems():
      print '%24s : %s' % (app, status)

  def _ConfigTest(self):
    print '# These are the commands that would be executed on start:'
    print ''

    for app in self._EnabledApps():
      print '# %s' % app
      print self._PrettyCommand(self._commands[app])
      print ''

  def _StopApps(self):
    for app in self._EnabledApps():
      attempts = 0
      signum = signal.SIGTERM
      self._logger.info('Stopping app: %s' % app)
      while True:
        status, pid = self._AppStatus(app)
        if status != STATUS_ALIVE:
          self._logger.info('%s stopped.' % app)
          break
        self._logger.info('Sending app %s signal %i' % (app, signum))
        os.kill(pid, signum)
        time.sleep(1.0)
        attempts += 1
        if attempts == 3:
          signum = signal.SIGKILL
        if attempts >= 6:
          self._logger.error('Too many attempts, aborting')
          break

  def _MainLoop(self):
    if len(self._extra_argv) != 2:
      self._Usage()
      sys.exit(1)
    command = self._extra_argv[1]

    if command == 'start':
      self._StartApps()
    elif command == 'stop':
      self._StopApps()
    elif command == 'status':
      self._Status()
    elif command == 'configtest':
      self._ConfigTest()
    else:
      self._Usage()
      sys.exit(1)


if __name__ == '__main__':
  KegbotMasterApp.BuildAndRun(name='master')



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

class Error(Exception):
  """An error occurred."""
  def Message(self):
    if self.message:
      return self.message
    m = self.__class__.__doc__
    m = m.split('\n', 1)[0]
    return m

class NotFoundError(Error):
  """The requested object could not be found."""

class ServerError(Error):
  """The server had a problem fulfilling your request."""

class BadRequestError(Error):
  """The request was incompleted or malformed."""

class NoAuthTokenError(Error):
  """An api_auth_token is required."""

class BadAuthTokenError(Error):
  """The api_auth_token given is invalid."""

class PermissionDeniedError(Error):
  """The api_auth_token given does not have permission for this resource."""

MAP_NAME_TO_EXCEPTION = dict((c.__name__, c) for c in Error.__subclasses__())

def ErrorCodeToException(code, message=None):
  cls = MAP_NAME_TO_EXCEPTION.get(code, Error)
  return cls(message)



from django import forms
from django.conf import settings

from registration.models import RegistrationProfile
from registration.forms import RegistrationForm

from pykeg.core import models

class KegbotRegistrationForm(RegistrationForm):
  WEIGHT_CHOICES = (
      (100, 'less than 100'),
      (120, '100-130'),
      (150, '131-170'),
      (180, '171+'),
  )
  GENDER_CHOICES = (
      ('male', 'Male'),
      ('female', 'Female'),
  )
  gender = forms.ChoiceField(choices=GENDER_CHOICES,
      help_text='Used for BAC estimation.')
  weight = forms.ChoiceField(choices=WEIGHT_CHOICES,
      help_text='Used for BAC estimation, kept private. You can lie.')
  twitter_name = forms.CharField(required=False,
      help_text='Do you use twitter? Enter your twitter name here to enable '
        'tweets announcing your drinks.')

  def save(self, profile_callback=None):
    new_user = RegistrationProfile.objects.create_inactive_user(username=self.cleaned_data['username'],
                                                                password=self.cleaned_data['password1'],
                                                                email=self.cleaned_data['email'],
                                                                send_email=False,
                                                                profile_callback=profile_callback)
    new_user.is_active = True
    new_user.save()
    new_profile, is_new = models.UserProfile.objects.get_or_create(user=new_user)
    new_profile.gender = self.cleaned_data['gender']
    new_profile.weight = self.cleaned_data['weight']
    new_profile.save()

    if 'pykeg.contrib.twitter' in settings.INSTALLED_APPS:
      twitter_name = self.cleaned_data.get('twitter_name')
      if twitter_name:
        from pykeg.contrib.twitter import models as twitter_models
        link = twitter_models.UserTwitterLink()
        link.user_profile = new_profile
        link.twitter_name = twitter_name
        link.save()

    return new_user


class UserProfileForm(forms.ModelForm):
  class Meta:
    model = models.UserProfile
    fields = ('gender', 'weight')


class MugshotForm(forms.Form):
  new_mugshot = forms.ImageField(required=False)


UNASSIGNED_TOKEN_QS = models.AuthenticationToken.objects.filter(user=None)
NEW_USERS = models.User.objects.all().order_by('-date_joined')

class ClaimTokenForm(forms.Form):
  token = forms.ModelChoiceField(queryset=UNASSIGNED_TOKEN_QS)
  user = forms.ModelChoiceField(queryset=NEW_USERS)



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import admin
from pykeg.billing import models

class CreditAdmin(admin.ModelAdmin):
  list_display = ('amount', 'acceptor', 'user', 'date')
  list_filter = ('user', 'acceptor', 'date')
  search_fields = ('user',)

admin.site.register(models.Credit, CreditAdmin)

class BillAcceptorAdmin(admin.ModelAdmin):
  list_display = ('name', 'increment')

admin.site.register(models.BillAcceptor, BillAcceptorAdmin)





# -*- coding: latin-1 -*-

from south.db import db
from django.db import models
from pykeg.contrib.soundserver.models import *

class Migration:
    
    def forwards(self, orm):
        
        # Adding model 'SoundEvent'
        db.create_table('soundserver_soundevent', (
            ('id', orm['soundserver.SoundEvent:id']),
            ('event_name', orm['soundserver.SoundEvent:event_name']),
            ('event_predicate', orm['soundserver.SoundEvent:event_predicate']),
            ('soundfile', orm['soundserver.SoundEvent:soundfile']),
            ('user', orm['soundserver.SoundEvent:user']),
        ))
        db.send_create_signal('soundserver', ['SoundEvent'])
        
        # Adding model 'SoundFile'
        db.create_table('soundserver_soundfile', (
            ('id', orm['soundserver.SoundFile:id']),
            ('sound', orm['soundserver.SoundFile:sound']),
            ('title', orm['soundserver.SoundFile:title']),
            ('active', orm['soundserver.SoundFile:active']),
        ))
        db.send_create_signal('soundserver', ['SoundFile'])
        
    
    
    def backwards(self, orm):
        
        # Deleting model 'SoundEvent'
        db.delete_table('soundserver_soundevent')
        
        # Deleting model 'SoundFile'
        db.delete_table('soundserver_soundfile')
        
    
    
    models = {
        'auth.group': {
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'unique_together': "(('content_type', 'codename'),)"},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'blank': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'blank': 'True'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'blank': 'True'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'contenttypes.contenttype': {
            'Meta': {'unique_together': "(('app_label', 'model'),)", 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'soundserver.soundevent': {
            'event_name': ('django.db.models.fields.CharField', [], {'max_length': '256'}),
            'event_predicate': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'soundfile': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['soundserver.SoundFile']"}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'soundserver.soundfile': {
            'active': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'sound': ('django.db.models.fields.files.ImageField', [], {'max_length': '100'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '128'})
        }
    }
    
    complete_apps = ['soundserver']



from pykeg.core import models

class KegbotSiteMiddleware:
  def process_request(self, request):
    if not hasattr(request, 'kbsite'):
      sitename = 'default'
      if 'site' in request.GET:
        sitename = request.GET['site']
      request.kbsite = models.KegbotSite.objects.get(name=sitename)



# -*- coding: latin-1 -*-

from south.db import db
from django.db import models
from pykeg.web.accounting.models import *

class Migration:
    
    def forwards(self, orm):
        
        # Deleting model 'drinkcharge'
        db.delete_table('accounting_drinkcharge')
        
        # Deleting model 'misccharge'
        db.delete_table('accounting_misccharge')
        
        # Deleting model 'billstatement'
        db.delete_table('accounting_billstatement')
        
        # Deleting model 'payment'
        db.delete_table('accounting_payment')
        
    
    
    def backwards(self, orm):
        
        # Adding model 'drinkcharge'
        db.create_table('accounting_drinkcharge', (
            ('base_funds', orm['accounting.payment:base_funds']),
            ('user', orm['accounting.payment:user']),
            ('statement', orm['accounting.payment:statement']),
            ('handling_funds', orm['accounting.payment:handling_funds']),
            ('ext_funds', orm['accounting.payment:ext_funds']),
            ('drink', orm['accounting.payment:drink']),
            ('id', orm['accounting.payment:id']),
            ('tax_funds', orm['accounting.payment:tax_funds']),
        ))
        db.send_create_signal('accounting', ['drinkcharge'])
        
        # Adding model 'misccharge'
        db.create_table('accounting_misccharge', (
            ('long_desc', orm['accounting.payment:long_desc']),
            ('base_funds', orm['accounting.payment:base_funds']),
            ('user', orm['accounting.payment:user']),
            ('statement', orm['accounting.payment:statement']),
            ('handling_funds', orm['accounting.payment:handling_funds']),
            ('ext_funds', orm['accounting.payment:ext_funds']),
            ('short_desc', orm['accounting.payment:short_desc']),
            ('id', orm['accounting.payment:id']),
            ('tax_funds', orm['accounting.payment:tax_funds']),
        ))
        db.send_create_signal('accounting', ['misccharge'])
        
        # Adding model 'billstatement'
        db.create_table('accounting_billstatement', (
            ('long_desc', orm['accounting.payment:long_desc']),
            ('statement_id', orm['accounting.payment:statement_id']),
            ('statement_date', orm['accounting.payment:statement_date']),
            ('short_desc', orm['accounting.payment:short_desc']),
            ('user', orm['accounting.payment:user']),
            ('id', orm['accounting.payment:id']),
        ))
        db.send_create_signal('accounting', ['billstatement'])
        
        # Adding model 'payment'
        db.create_table('accounting_payment', (
            ('base_funds', orm['accounting.payment:base_funds']),
            ('payment_status', orm['accounting.payment:payment_status']),
            ('tax_funds', orm['accounting.payment:tax_funds']),
            ('handling_funds', orm['accounting.payment:handling_funds']),
            ('user', orm['accounting.payment:user']),
            ('id', orm['accounting.payment:id']),
            ('payment_date', orm['accounting.payment:payment_date']),
            ('source_user', orm['accounting.payment:source_user']),
            ('statement', orm['accounting.payment:statement']),
            ('ext_funds', orm['accounting.payment:ext_funds']),
            ('status_message', orm['accounting.payment:status_message']),
            ('payment_funds_source', orm['accounting.payment:payment_funds_source']),
            ('transaction_id', orm['accounting.payment:transaction_id']),
        ))
        db.send_create_signal('accounting', ['payment'])
        
    
    
    models = {
        
    }
    
    complete_apps = ['accounting']



# Copyright 2008-2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegnet client/server implementation."""

# TODO(mikey): need to isolate internal-only events (like QuitEvent) from
# external ones (like FlowUpdate).
# TODO(mikey): add onDisconnect handler and retry/backoff
# TODO(mikey): also raise an exception on socket errors

import asyncore
import asynchat
import cStringIO
import logging
import Queue
import socket
import struct
import sys
import threading
import time

# Use simplejson if python version is < 2.6
if sys.version_info[:2] < (2, 6):
  import simplejson as json
else:
  import json

import gflags

from pykeg.core import kb_common
from pykeg.core import protoutil
from pykeg.core import util
from pykeg.core.net import kegnet_pb2
from google.protobuf import message as protobuf_message

FLAGS = gflags.FLAGS

gflags.DEFINE_string('kb_core_addr', kb_common.KB_CORE_DEFAULT_ADDR,
    'The address of the kb_core host, when connecting as a client. '
    'Specify as "<hostname>:<port>".')

gflags.DEFINE_string('kb_core_bind_addr', kb_common.KB_CORE_DEFAULT_ADDR,
    'Address that the kegnet server should bind to. '
    'Specify as "<hostname>:<port>".')

gflags.DEFINE_string('tap_name', 'kegboard.flow0',
    'Default tap name.')

MESSAGE_TERMINATOR = '\n\n'

def ProtoMessageToJson(message):
  wrapper = {
    'method': message.DESCRIPTOR.name,
    'id': None,
    'params': protoutil.ProtoMessageToDict(message),
  }
  return json.dumps(wrapper)

NAME_TO_MESSAGE_CLASS = {}
for cls in protobuf_message.Message.__subclasses__():
  NAME_TO_MESSAGE_CLASS[cls.DESCRIPTOR.name] = cls

def MessageDictToProtoMessage(message_dict):
  message_name = message_dict['method']
  message_class = NAME_TO_MESSAGE_CLASS[message_name]
  return protoutil.DictToProtoMessage(message_dict['params'], message_class())

def ProtoMessageToShortStr(message):
  name = message.DESCRIPTOR.name
  values = str(message).replace('\n', ' ')
  # TODO(mikey): show submessages properly
  return '<%s: %s>' % (name, values)

class KegnetProtocolHandler(asynchat.async_chat):
  """A general purpose request handler for the Kegnet protocol.

  This async_chat subclass can be used for client and server implementations.
  The handler will call _HandleKegnetMessage on receipt of a complete kegnet
  message.
  """
  def __init__(self, sock=None):
    asynchat.async_chat.__init__(self, sock)
    self.set_terminator(MESSAGE_TERMINATOR)
    self._ibuffer = cStringIO.StringIO()
    self._logger = logging.getLogger('kegnet')
    self._in_notifications = Queue.Queue()
    self._lock = threading.Lock()
    self._quit = False

  def stop(self):
    self._quit = True
    if self.socket:
      self.close()

  ### async_chat methods
  @util.synchronized
  def initiate_send(self):
    return asynchat.async_chat.initiate_send(self)

  def collect_incoming_data(self, data):
    self._ibuffer.write(data)

  def found_terminator(self):
    strbuf = self._ibuffer.getvalue()
    self._ibuffer.seek(0)
    self._ibuffer.truncate()

    if not strbuf:
      self._logger.warning('Received empty message')
      return

    try:
      message_dict = json.loads(strbuf)
    except ValueError:
      self._logger.warning('Received malformed message, dropping.')
      return

    # Distinguish between requests, responses, and notifications, using JSON-RPC
    # v2.0 rules
    # (http://groups.google.com/group/json-rpc/web/json-rpc-1-2-proposal):

    self._logger.debug('Received message: %s' % message_dict)

    if 'method' in message_dict and 'params' in message_dict:
      # Requests and Notifications both have 'method' and 'param' fields;
      # notifications have the 'id' field set to null (None).
      if 'id' in message_dict and message_dict['id'] is not None:
        # Message is a request.
        self.HandleRequest(message_dict)
      else:
        # Message is a notification.
        self.HandleNotification(message_dict)
    elif 'result' in message_dict:
      self.HandleResponse(message_dict)
    else:
      self._logger.warning('Received unknown message type, dropping.')

  def handle_close(self):
    asynchat.async_chat.handle_close(self)
    #self._server.ChannelClosed(self)

  ### KegnetProtocolHandler methods
  def HandleNotification(self, message_dict):
    self._logger.debug('Received notification: %s' % message_dict)
    message = MessageDictToProtoMessage(message_dict)
    self._in_notifications.put(message)

  def HandleRequest(self, message_dict):
    pass

  def HandleResponse(self, message_dict):
    pass

  def PopNotification(self, timeout=None):
    return self._in_notifications.get(timeout=timeout)

  def SendMessage(self, msg):
    str_message = ProtoMessageToJson(msg)
    self._logger.debug('Sending JSON message: %s' % str_message)
    self.push(str_message + MESSAGE_TERMINATOR)


class KegnetServerHandler(KegnetProtocolHandler):
  """ An asyncore handler for the core kegnet server. """
  def __init__(self, sock, server):
    KegnetProtocolHandler.__init__(self, sock)
    self._server = server
    self._server.ChannelOpened(self)

  def handle_close(self):
    KegnetProtocolHandler.handle_close(self)
    self._logger.info('Closing down...')
    self._server.ChannelClosed(self)

  def _HandleBadMessage(self, strdata, exception):
    self._server._logger.warn("Unknown command from %s, closing connection" %
                              str(self.addr))
    self._server._logger.warn("Exception was: %s" % str(exception))
    self._server.ChannelClosed(self)
    self.close()

  def HandleNotification(self, message_dict):
    KegnetProtocolHandler.HandleNotification(self, message_dict)
    event_hub = self._server._kb_env.GetEventHub()
    event_hub.PublishEvent(self.PopNotification())


class KegnetClient(KegnetProtocolHandler):
  RECONNECT_BACKOFF = [5, 5, 10, 10, 20, 20, 60]
  def __init__(self, addr=None):
    if not addr:
      addr = FLAGS.kb_core_addr
    self._addr = util.str_to_addr(addr)
    self._last_reconnect = 0
    self._num_retries = 0
    KegnetProtocolHandler.__init__(self)

  def Reconnect(self, force=False):
    backoff_secs = self._ReconnectTimeout()
    if backoff_secs:
      time.sleep(backoff_secs)
    self._last_reconnect = time.time()

    if self.connected:
      return True

    self._logger.info('Connecting to %s:%s' % self._addr)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    try:
      sock.connect(self._addr)
      self.set_socket(sock)
      self.onConnected()
      self._num_retries = 0
      return True
    except socket.error:
      self._num_retries += 1
      self._logger.info('Connection failed. Retry in %i seconds.' %
          backoff_secs)
      return False

  def _ReconnectTimeout(self):
    if not self._last_reconnect:
      return 0
    prev_wait = time.time() - self._last_reconnect
    pos = min(self._num_retries, len(self.RECONNECT_BACKOFF) - 1)
    amt = self.RECONNECT_BACKOFF[pos]
    if prev_wait < amt:
      return amt - prev_wait
    return 0

  ### convenience functions
  def SendPing(self):
    message = kegnet_pb2.Ping()
    return self.SendMessage(message)

  def SendMeterUpdate(self, tap_name, meter_reading):
    message = kegnet_pb2.MeterUpdate()
    message.tap_name = tap_name
    message.reading = meter_reading
    return self.SendMessage(message)

  def SendFlowStart(self, tap_name):
    message = kegnet_pb2.FlowRequest()
    message.tap_name = tap_name
    message.request = message.START_FLOW
    return self.SendMessage(message)

  def SendFlowStop(self, tap_name):
    message = kegnet_pb2.FlowRequest()
    message.tap_name = tap_name
    message.request = message.STOP_FLOW
    return self.SendMessage(message)

  def SendThermoUpdate(self, sensor_name, sensor_value):
    message = kegnet_pb2.ThermoEvent()
    message.sensor_name = sensor_name
    message.sensor_value = sensor_value
    return self.SendMessage(message)

  def SendAuthTokenAdd(self, tap_name, auth_device_name, token_value):
    message = kegnet_pb2.TokenAuthEvent()
    message.tap_name = tap_name
    message.auth_device_name = auth_device_name
    message.token_value = token_value
    message.status = message.ADDED
    return self.SendMessage(message)

  def SendAuthTokenRemove(self, tap_name, auth_device_name, token_value):
    message = kegnet_pb2.TokenAuthEvent()
    message.tap_name = tap_name
    message.auth_device_name = auth_device_name
    message.token_value = token_value
    message.status = message.REMOVED
    return self.SendMessage(message)

  def onConnected(self):
    self._logger.info('Connected!')

  def onDisconnected(self):
    self._logger.info('Disconnected!')


class SimpleKegnetClient(KegnetClient):

  def serve_forever(self):
    self.Reconnect()
    while not self._quit:
      if not asyncore.socket_map:
        self.onDisconnected()
        self.Reconnect()
        continue
      asyncore.loop(timeout=0.5, count=1)

  def HandleNotification(self, message_dict):
    self._logger.debug('Received notification: %s' % message_dict)
    event = MessageDictToProtoMessage(message_dict)
    if isinstance(event, kegnet_pb2.FlowUpdate):
      self.onFlowUpdate(event)
    elif isinstance(event, kegnet_pb2.DrinkCreatedEvent):
      self.onDrinkCreated(event)
    elif isinstance(event, kegnet_pb2.CreditAddedEvent):
      self.onCreditAdded(event)

  def onFlowUpdate(self, event):
    pass

  def onDrinkCreated(self, event):
    pass

  def onCreditAdded(self, event):
    pass


class KegnetClientThread(util.KegbotThread):
  def __init__(self, name, client):
    util.KegbotThread.__init__(self, name)
    self._client = client

  def run(self):
    self._client.serve_forever()


class KegnetServer(asyncore.dispatcher):
  """asyncore server implementation for Kegnet protocol"""
  def __init__(self, name, kb_env, addr='', port=0, qsize=5):
    self._name = name
    self._kb_env = kb_env
    self._logger = logging.getLogger(self._name)
    self._bind_address = util.str_to_addr(addr)
    self._qsize = qsize
    self._clients = set()
    self._lock = threading.Lock()
    asyncore.dispatcher.__init__(self)

  def StartServer(self):
    self._logger.info("Starting server on %s" % str(self._bind_address,))
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(self._bind_address)
    self.listen(self._qsize)

  def StopServer(self):
    self._logger.info("Stopping server")
    self.close()

  @util.synchronized
  def ChannelOpened(self, channel):
    self._logger.info('Remote host connected: %s:%i' % channel.addr)
    self._clients.add(channel)

  @util.synchronized
  def ChannelClosed(self, channel):
    self._logger.info('Remote host closed: %s:%i' % channel.addr)
    self._clients.remove(channel)

  @util.synchronized
  def SendEventToClients(self, event):
    # TODO(mikey): filter events -- should be based on subscriptions & exclude
    # internal events.
    if not self._clients:
      return
    self._logger.info('Sending event to %i client(s): %s' %
      (len(self._clients), ProtoMessageToShortStr(event)))
    str_message = ProtoMessageToJson(event)
    for client in self._clients:
      try:
        client.push(str_message + MESSAGE_TERMINATOR)
      except IndexError:
        self._logger.warning('Exception sending to %s' % client)
        util.LogTraceback(self._logger.warning)

  ### asyncore.dispatcher methods

  def handle_accept(self):
    # TODO(mikey): error handling
    conn, addr = self.accept()
    conn.settimeout(1)
    KegnetServerHandler(conn, self)


if __name__ == '__main__':
  server = KegnetServer(name='kegnet', kb_env=None,
      addr=FLAGS.kb_core_bind_addr)

  print "Start asyncore"
  asyncore.loop(timeout=2)
  print "finished loop"
